/* ident "@(#)tftpd.c 1.52, dirk" */

/*
 * TFTPD.C - TFTPD daemon
 *
 * COPYRIGHT (c) 1992,96 Dirk Koeppen EDV-Beratungs-GmbH, W-63073 Offenbach
 *
 * delta date      modification					done by
 * ----- --------- -------------------------------------------- -------------
 *  1.00 1992      initial version				dirk
 */

#pragma warning( disable : 4201 4127 )		// disable:  4201 nameless struct/union warnings from winnt.h
											//			 4127 conditional expression is constant (from _ASSERT()s)
//AVL #define _DEBUG 1
#include <sys/types.h>
#include <ctype.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/stat.h>
#ifndef UNIX
# include <conio.h>
#endif
#include <stdlib.h>
#include <stdarg.h>
#include <stddef.h>
#include <time.h>
#include <string.h>
#ifndef UNIX
# include <direct.h>
#endif
#include <errno.h>
#if defined(WNAPP) || defined(NTAPP) || defined(NTSVC)
# ifdef WNAPP
#  include "wsock.h"
# endif 
# if defined(NTAPP) || defined(NTSVC) 
#  include "wsock32.h"
# endif
# include <stdarg.h>
# include <stdlib.h>
# include <string.h>
# include <sys/stat.h>
# include <time.h>
# include <io.h>
# include <direct.h>
#else
#ifdef OS2
# define BSD_S
# include <sys/select.h>
# include <utils.h>
#endif
# include <sys/socket.h>
# include <sys/time.h>
# include <sys/ioctl.h>
# include <netinet/in.h>
# ifndef UNIX
#  include <process.h>
# endif
# include <netdb.h>
#endif

#ifdef UNIX
# include "unix.h"
#endif

#include "tftp.h"
#include "tftpd.ver"
#include "tftpdmsg.h"

#ifdef _DEBUG
#include <crtdbg.h>
#include <tchar.h>
#include "service.h"
#include "debug.h"
#endif	/* _DEBUG */

#ifdef IP_MULTICAST_TTL
#undef IP_MULTICAST_TTL
#endif
#define IP_MULTICAST_TTL 10


#ifdef NW
extern int	errno;
#endif
#ifndef LINUX
extern char 	*sys_errlist[];
#endif

#ifdef UNIX
typedef int LONG;
typedef int DWORD;
#define FALSE 0
#define stricmp strcasecmp
#endif

/* tuneables... */
					/* temp. buffer size */
#define	PKTSIZE			(MSEGSIZE+16)
#define	DEF_TFTPDPORT		69	/* default port if no services entry */
#define	DEF_TFTPDAHEAD		0	/* default do not enable read-ahead */
#define	DEF_TFTPDTMOUTVAL	15	/* def. timeout val in secs */
#define	DEF_TFTPDRESTRICT	0	/* def. file restriction is no */
#define TFTPDMAXTRANS		64	/* # of concur. TFTP transfers */
#define DEF_TFTPDMAXTRANS	32	/* def. # of concur. TFTP transfers */
#define	DEF_TFTPDVERBOSE	1	/* def. verbosity */
#define	DEF_TFTPDUNICAST	1	/* def. number of unicast to send */
#define MTFTPDMAXFILES		16	/* max. number of MTFTP files */
#define	LAST_RETRY		3	/* max. number of retransmit/pack. */
#define	MAX_RETRY		32	/* max. number of retransmit */

#ifdef NW
					/* def. directory to change to */
# define	DEF_TFTPDCHDIR		"sys:\\tftpboot"
					/* def. file to log to */
# define	DEF_TFTPDLOGFIL		"sys:\\tftpboot\\tftpd.log"
#else
# if defined(WNAPP) || defined(NTAPP) || defined(NTSVC) || defined(OS2)
					/* def. directory to change to */
#  define	DEF_TFTPDCHDIR		"c:\\tftpboot"
					/* def. file to log to */
#  define	DEF_TFTPDLOGFIL		"c:\\tftpboot\\tftpd.log"
					/* def. file for MTFTP configuration */
#  define	SZSERVICENAME		DESCRIPTION // _T("Intel LANDesk TFTPD Service") // "TFTPD"
					/* def. file for TFTP configuration */
#  define	DEF_TFTPTAB			"c:\\etc\\tftpd.cnf"
# else
					/* def. directory to change to */
#  define	DEF_TFTPDCHDIR		"/tftpboot"
					/* def. file to log to */
#  define	DEF_TFTPDLOGFIL		"/tftpboot/tftpd.log"
# endif
#endif
/* end of tuneable section */


/* some shorties... */
#define	SA	struct sockaddr
#define	SI	struct sockaddr_in
#define	TH	struct tftphdr

/* global data... */
int	TFTPDMaxTrans,		/* max. number of active TFTP transfers */
	TFTPDSocket,		/* TFTPD connection socket */
	LTFTPDSocket,		/* LTFTPD connection socket */
	MTFTPDSocket,		/* MTFTPD connection socket */
	TFTPDTmoutval,		/* timeout value in secs */
	TFTPDTimeout,		/* max. seconds to wait until TFTPD exit */
	TFTPDVerbose,		/* verbose option at program start */
	LTFTPDSegsize,		/* segment size */
	TFTPDUnicast,		/* number of unicast packages */
	TFTPDAhead,		/* enable read-ahead */
	TFTPDWrite,		/* enable TFTP writes */
	TFTPDCreat,		/* allow to create files */
	TFTPDRestrict,		/* allow only file from current directory */
	LTFTPDFlag,		/* flag to use LTFTPD */
	MTFTPDFlag;		/* flag to use MTFTPD */
char	*TFTPDChdir,		/* directory to change to */
	*TFTPDLogFil,		/* filename where to log to */
	*TFTPDMconfig,		/* MTFTP configuration file */
	Buf[MSEGSIZE*2];	/* general buffer */
FILE	*TFTPDLogFp;		/* fp where to log to */
unsigned short	TFTPDPort,	/* TFTP port */
	MTFTPDPort,		/* MTFTP port */
	MTFTPPort,		/* MTFTP port */
	LTFTPDPort;		/* large TFTP port */

#ifdef NTSVC
HKEY	hTftpdKey = NULL;
HKEY	hTftpdFilesKey = NULL;
HANDLE fhand;
#endif
int McastSegSize = DEFAULT_SEG_SIZE;
char FileName[256], Base_Path[256];
#ifdef AVL
FILE	*AVLFp;		/* fp where to log to */
#endif

#ifdef _DEBUG
static HANDLE hOutConsole;
TCHAR szDebugBuffer[512];
DWORD dwDebug;
#endif

#if defined(WNAPP) || defined(NTSVC)
int	doclose;		/* close this daemon */
#endif

#if defined(NTAPP) || defined(NTSVC)
HWND 	hwndMain;        	/* main window handle */
HDC		Hdc;
HINSTANCE hInst;      		/* instance handle */
long	Tcnt;				/* timeout counter */
#define	CHAR_WIDTH	9		/* character width */
#define CHAR_HIGHT	15		/* ...and height */
#define	XLEN		132		/* default width and length of the window */
#define	VXLEN		70		/* visiable length */
#define	YLEN		24
int		YLine;				/* text output position */
char	Screen[YLEN][XLEN+1]; /* screen buffer */

extern int 		_argc;
extern char 	**_argv;
#endif

#ifdef NTSVC
BOOL					StopErrno=0;
SERVICE_STATUS			ssStatus;		/* current status of the service */
SERVICE_STATUS_HANDLE	sshStatusHandle;
char					szErr[512];
DWORD					dwErr;
int						NTsvc;			/* flag that this is running as NT service */
char					Cmdline[256];	/* command line from tftptab */
#endif

struct _errmsg {		/* TFTP error codes */
	int	e_code;
	char	*e_msg;
} TFTPDErrmsgs[]= {
	{ EUNDEF,	"Undefined error code" },
	{ ENOTFOUND,	"File not found" },
	{ EACCESS,	"Access violation" },
	{ ENOSPACE,	"Disk full or allocation exceeded" },
	{ EBADOP,	"Illegal TFTP operation" },
	{ EBADID,	"Unknown transfer ID" },
	{ EEXISTS,	"File already exists" },
	{ ENOUSER,	"No such user" },
	{ EMCASTADDR,	"No multicast address" },
	{ -1,		0 }
};

#ifdef AVL
struct _mtftpc {		/* MTFTP configurations */
	char	fnam[80];
	long	addr;
	int	port;
} Mtftpc[MTFTPDMAXFILES+1];
#endif	/* AVL */

struct _trans {			/* TFTP tasks */
	int	fd;
	SI	si;
	SI	msi;
	time_t	first, last;	/* 0 if free */
	int	retry;
	int	lpack;
	int	done;
	long	fsz;
	int	ucnt;
	char	*abuf;
	int	alen;
	int	apack;
	int segsize;		/* different seg sizes negotiated by different connections */
	char *fname;	/* will be allocated only incase of mcast */
} Trans[TFTPDMAXTRANS];


struct _files {	/* open files */
	int	use;
	int	fd;
	char	fn[80];
} Files[TFTPDMAXTRANS];

#define SZDEPENDENCIES       "Tcpip\0\0"

/* LOG0 - log fatal errors
 * VARARGS1 PRINTFLIKE1 */
void
log0(char *fmt, ...)
{
	va_list	args;
	char	ln[256];
#if defined(NTAPP) || defined(NTSVC)
	int		ypos, len;
#endif
#ifdef NTSVC
	HANDLE	hEventSource;
	LPTSTR	lpszStrings[2];
#endif

	/* print something to stderr */
	va_start(args, fmt);

	vsprintf (ln, fmt, args);

#ifdef NW
	ConsolePrintf ("%s", ln);
#elif defined(NTAPP) || defined(NTSVC)
# ifdef NTSVC
	if (!NTsvc) {
# endif
		/* print to windows and store a copy */
		ypos= YLine % YLEN;
		len= strlen (ln);
	
		memcpy (Screen[ypos], ln, len);
		if (Screen[ypos][len-1] == '\n')
			len--;
		memset (Screen[ypos]+len, ' ', XLEN-len);

		TextOut (Hdc, CHAR_WIDTH/2, CHAR_HIGHT/2 + (YLine % YLEN)
			* CHAR_HIGHT, Screen[ypos], XLEN);

		ypos= ++YLine % YLEN;
		memset (Screen[ypos], ' ', XLEN);

		TextOut (Hdc, CHAR_WIDTH/2, CHAR_HIGHT/2 + (YLine % YLEN)
			* CHAR_HIGHT, Screen[ypos], XLEN);
# ifdef NTSVC
	}
	else {
		/* use event logging */
		hEventSource= RegisterEventSource (NULL, SZSERVICENAME);
	
		if (ln[strlen(ln)-1] == '\n')
			ln[strlen(ln)-1]= 0;
		lpszStrings[0]= ln;
		lpszStrings[1]= 0;
	
		if (hEventSource != NULL) {
			ReportEvent (
				hEventSource,
				EVENTLOG_ERROR_TYPE,
				0,
				TFTPD_EVENTID,
				NULL,
				1,
				0,
				lpszStrings,
				NULL);
		
			DeregisterEventSource (hEventSource);
		}
	}
# endif
#else
	/* simply print */
	printf ("%s", ln);
#endif

	if (TFTPDLogFp)
		fprintf (TFTPDLogFp, "%s", ln);

	va_end(args);
}


/* LOG1 - log errors 
 * VARARGS1 PRINTFLIKE1 */
void
log1(char *fmt, ...)
{
	va_list	args;
	char	ln[256];
#if defined(NTAPP) || defined(NTSVC)
	int		ypos, len;
#endif
#ifdef NTSVC
	HANDLE	hEventSource;
	LPTSTR	lpszStrings[2];
#endif

	/* print something to stderr */
	va_start(args, fmt);

	if (TFTPDVerbose >= 1) {
		vsprintf (ln, fmt, args);

#ifdef WNAPP
		printf ("%s", ln);
#elif defined(NTAPP) || defined(NTSVC)
# ifdef NTSVC
		if (!NTsvc) {
# endif
			/* print to windows and store a copy */
			ypos= YLine % YLEN;
			len= strlen (ln);
	
			memcpy (Screen[ypos], ln, len);
			if (Screen[ypos][len-1] == '\n')
				len--;
			memset (Screen[ypos]+len, ' ', XLEN-len);

			TextOut (Hdc, CHAR_WIDTH/2, CHAR_HIGHT/2 + (YLine % YLEN)
				* CHAR_HIGHT, Screen[ypos], XLEN);

			ypos= ++YLine % YLEN;
			memset (Screen[ypos], ' ', XLEN);

			TextOut (Hdc, CHAR_WIDTH/2, CHAR_HIGHT/2 + (YLine % YLEN)
				* CHAR_HIGHT, Screen[ypos], XLEN);
# ifdef NTSVC
		}
		else {
			/* use event logging */
			hEventSource= RegisterEventSource (NULL, SZSERVICENAME);
	
			if (ln[strlen(ln)-1] == '\n')
				ln[strlen(ln)-1]= 0;
			lpszStrings[0]= ln;
			lpszStrings[1]= 0;
	
			if (hEventSource != NULL) {
				ReportEvent (
					hEventSource,
					EVENTLOG_WARNING_TYPE,
					0,
					TFTPD_EVENTID,
					NULL,
					1,
					0,
					lpszStrings,
					NULL);
		
				DeregisterEventSource (hEventSource);
			}
		}
# endif
#else
		/* simply print */
		fprintf (stderr, "%s", ln);
		fflush (stderr);
#endif

		if (TFTPDLogFp)
			fprintf (TFTPDLogFp, "%s", ln);
	}

	va_end(args);
}


/* LOG2 - log verbose messages 
 * VARARGS1 PRINTFLIKE1 */
void
log2(char *fmt, ...)
{
	va_list	args;
	char	ln[256];
#if defined(NTAPP) || defined(NTSVC)
	int		ypos, len;
#endif
#ifdef NTSVC
	HANDLE	hEventSource;
	LPTSTR	lpszStrings[2];
#endif

	/* print something to stderr */
	va_start(args, fmt);

	if (TFTPDVerbose >= 2) {
		vsprintf (ln, fmt, args);

#ifdef WNAPP
		printf ("%s", ln);
#elif defined(NTAPP) || defined(NTSVC)
# ifdef NTSVC
		if (!NTsvc) {
# endif
			/* print to windows and store a copy */
			ypos= YLine % YLEN;
			len= strlen (ln);
	
			memcpy (Screen[ypos], ln, len);
			if (Screen[ypos][len-1] == '\n')
				len--;
			memset (Screen[ypos]+len, ' ', XLEN-len);

			TextOut (Hdc, CHAR_WIDTH/2, CHAR_HIGHT/2 + (YLine % YLEN)
				* CHAR_HIGHT, Screen[ypos], XLEN);

			ypos= ++YLine % YLEN;
			memset (Screen[ypos], ' ', XLEN);

			TextOut (Hdc, CHAR_WIDTH/2, CHAR_HIGHT/2 + (YLine % YLEN)
				* CHAR_HIGHT, Screen[ypos], XLEN);
# ifdef NTSVC
		}
		else {
			/* use event logging */
			hEventSource= RegisterEventSource (NULL, SZSERVICENAME);
	
			if (ln[strlen(ln)-1] == '\n')
				ln[strlen(ln)-1]= 0;
			lpszStrings[0]= ln;
			lpszStrings[1]= 0;
	
			if (hEventSource != NULL) {
				ReportEvent (
					hEventSource,
					EVENTLOG_INFORMATION_TYPE,
					0,
					TFTPD_EVENTID,
					NULL,
					1,
					0,
					lpszStrings,
					NULL);
		
				DeregisterEventSource (hEventSource);
			}
		}
# endif
#else
		/* simply print */
		fprintf (stderr, "%s", ln);
		fflush (stderr);
#endif

		if (TFTPDLogFp)
			fprintf (TFTPDLogFp, "%s", ln);
	}

	va_end(args);
}


/* MOPEN - my open which shares filedes */
int
mopen (char *fn, int mode)
{
	int	i;

	/* check if open for writing */
	if (mode == P_WRQ) {
		for (i= 0; i < TFTPDMAXTRANS; i++)
			if (Files[i].use == 0) {
				if ((Files[i].fd= open (fn,
					O_WRONLY|O_BINARY|O_CREAT|O_TRUNC, S_IREAD | S_IWRITE)) < 0)
					return (-1);
				Files[i].use= 1;
				strncpy (Files[i].fn, fn, sizeof (Files[i].fn));
				return (Files[i].fd);
			}

		return (-1);
	}

	/* no, it's for reading */
	for (i= 0; i < TFTPDMAXTRANS; i++) {
		if (Files[i].use && !strcmp (fn, Files[i].fn)) {
			Files[i].use++;
			return (Files[i].fd);
		}
	}

	for (i= 0; i < TFTPDMAXTRANS; i++)
		if (Files[i].use == 0) {
			if ((Files[i].fd= open (fn, O_RDONLY|O_BINARY)) < 0)
				return (-1);
			Files[i].use= 1;
			strncpy (Files[i].fn, fn, sizeof (Files[i].fn));
			return (Files[i].fd);
		}

	return (-1);
}


/* MCLOSE - my close which shares fildes */
int
mclose(int fd)
{
	int	i, r;

	for (i= 0; i < TFTPDMAXTRANS; i++)
		if (Files[i].use && Files[i].fd == fd) {
			Files[i].use--;
			r= 0;
			if (Files[i].use == 0) {
				if ((r= close (Files[i].fd)) < 0) {
					Files[i].use = 0;
					return (r);
				}
			}
			return (r);
		}

	return (-1);
}


/* USAGE - show the NLM usage */
void
usage(void)
{
#if defined(NTAPP) || defined(NTSVC) 
	MSG	msg;
#endif
#ifdef NW
	/* remove NLM screen */
	DestroyScreen (GetCurrentScreen ());
#endif

	log0("TFTPD: illegal usage\n");
#ifndef NTSVC
	log0("usage: TFTPD [-c num] [-d [path]] [-h] [-i num] [-k num] [-l [file]] \\\n");
	log0("             [-m file port] [-p num] [-r] [-s num port] [-u num] [-v 0-2] [-w [-x]]\n");
#else
	log0("usage: TFTPD -cmd [options]       (run NT service directly on console)\n");
	log0("       TFTPD -install             (install as NT service)\n");
	log0("       TFTPD -remove              (remove from NT service database)\n");
#endif
	log0("where: -c num         (number of concurrent TFTP transfers max. %d)\n",
			TFTPDMAXTRANS);
	log0("       -d [path]      (specify the path to chdir)\n");
	log0("       -h             (enable read ahead buffers)\n");
	log0("       -i num         (exit after num seconds of inactivity, 0=never)\n");
	log0("       -k num         (set TFTP transmission keep alive, def. %d)\n", 
		DEF_TFTPDTMOUTVAL);
	log0("       -l [file]      (log all messages to file)\n");
	log0("       -m file port   (MTFTP configuration file)\n");
	log0("       -p num         (set TFTPD listen port)\n");
	log0("       -r             (restrict file access to current dir)\n");
	log0("       -s num port    (set TFTP segment size max. %d)\n", MSEGSIZE);
	log0("       -u num         (MTFTP unicasts to be send def. %d)\n",
		DEF_TFTPDUNICAST);
	log0("       -v 0-2         (verbosity 0-less, 2-most)\n");
	log0("       -w             (enable TFTP writes)\n");
	log0("       -x             (allow to create files)\n");

#if defined(NTAPP) || defined(NTSVC)
	/* loop */
# ifdef NTSVC
	if (!NTsvc) {
# endif
		while (GetMessage (&msg, NULL, 0, 0)) {
			TranslateMessage (&msg);
			DispatchMessage (&msg);
		}
# ifdef NTSVC
	}
# endif
#endif

	exit (1);
}


/* TFTPDATEXIT - close resources */
void
TFTPDatexit(void)
{
	int	i;

#ifdef WNAPP
	log0("TFTPD: TFTP daemon is shut down\n");
	doclose= 1;
#endif
#if defined(WNAPP) || defined(NTAPP) || defined(NTSVC)
//AVL	WSACancelBlockingCall ();	// problem with wsock 2.0
	closesocket(TFTPDSocket);
	if (LTFTPDSocket >= 0)
		closesocket(LTFTPDSocket);
	if (MTFTPDSocket >= 0)
		closesocket(MTFTPDSocket);

	/* release WinSock DLL */
    WSACleanup();
#else
	close (TFTPDSocket);
	if (LTFTPDSocket >= 0)
		close (LTFTPDSocket);
	if (MTFTPDSocket >= 0)
		close (MTFTPDSocket);
#endif

	for (i= 0; i < TFTPDMaxTrans; i++)
		if (Trans[i].first != 0 && Trans[i].fd >= 0 &&
			Trans[i].lpack == Trans[i].lpack) {
			mclose (Trans[i].fd);
			if (Trans[i].abuf)
				free(Trans[i].abuf);
			Trans[i].abuf= 0;
			Trans[i].fd= -1;

			if (Trans[i].fname) 
				free(Trans[i].fname);
			Trans[i].fname = (char *) 0;
		}
	if (TFTPDLogFp != NULL)
		fclose (TFTPDLogFp);
}


/* ACK - send a ack packet (positive acknowledge) */
void
ack(int s, SI *to, int tolen, int pnum)
{
	TH		*tp;
	int 		r, length;

	/* setup the error package */
	tp= (TH *)Buf;
	tp->th_opcode= htons(P_ACK);
	tp->th_code= htons((unsigned short)pnum);
	length= sizeof(tp->th_opcode) + sizeof(tp->th_code);

	/* send it to the client, if to == 0 the socket is already bound */
	if (to == 0)
		r= send (s, (char *)tp, length, 0);
	else
		r= sendto (s, (char *)tp, length, 0, (SA *)to, tolen);
	if (r != length)
		log1("TFTPD[sendto(ack)]: errno= %d\n", errno);
}


/* NAK - send a nak packet (error message)
 *	error code passed in is one of the standard TFTP codes
 *	or an UNIX errno offset by 100. if to == 0 socket is bound */
void
nak(int s, SI *to, int tolen, int error)
{
	TH		*tp;
	int 		r, length;
	struct _errmsg	*pe;

	/* setup the error package */
	tp= (TH *)Buf;
	tp->th_opcode= htons(P_ERROR);
	tp->th_code= htons((short)error);

	/* search for the error string */
	for (pe= TFTPDErrmsgs; pe->e_code >= 0; pe++)
		if (pe->e_code == error)
			break;
	if (pe->e_code < 0) {
		pe->e_msg= sys_errlist[error - 100];
		tp->th_code= htons(EUNDEF);   
	}

	/* copy the error description */
	strcpy (tp->th_msg, pe->e_msg);
	length= strlen(pe->e_msg);
	tp->th_msg[length]= 0;
	length+= 5;
	
	/* send it to the client, if to == 0 the socket is already bound */
	if (to == 0)
		r= send (s, (char *)tp, length, 0);
	else
		r= sendto (s, (char *)tp, length, 0, (SA *)to, tolen);
	if (r != length)
		log1("TFTPD[sendto(nak)]: errno= %d\n", errno);
}

/* OACK - send an option ack packet (positive acknowledge) */
void
oack(int s, SI *to, int tolen, long fsize, int bsize)
{
	TH		*tp;
	int 		r, length;
	char OackBuf[32];

	/* setup the error package */
	tp= (TH *)OackBuf;
	tp->th_opcode= htons(P_OACK);
	/* fill tsize and blksize options */
	r = 0;
	if (fsize)
	{
		strcpy(&tp->th_stuff[r], "tsize");
		r += 6;	/* sizeof "tsize" with zero */
#ifdef NTSVC
		_ltoa(fsize, &tp->th_stuff[r], 10);
#else
		sprintf(&tp->th_stuff[r],"%d",fsize);
#endif
		r += strlen(&tp->th_stuff[r]) + 1;
	}
	if (bsize)
	{
		strcpy(&tp->th_stuff[r], "blksize");
		r += 8;	/* sizeof "blksize" */
#ifdef NTSVC
		_itoa(bsize, &tp->th_stuff[r], 10);
#else
		sprintf(&tp->th_stuff[r],"%d",bsize);
#endif
		r += strlen(&tp->th_stuff[r]) + 1;
	}

/* 
	length= sizeof(tp->th_opcode) + sizeof(tp->th_stuff);
*/
	length= sizeof(tp->th_opcode) + r;

	/* send it to the client, if to == 0 the socket is already bound */
	if (to == 0)
		r= send (s, (char *)tp, length, 0);
	else
		r= sendto (s, (char *)tp, length, 0, (SA *)to, tolen);
	if (r != length)
		log1("TFTPD[sendto(ack)]: errno= %d\n", errno);
}


/* VACCESS - validate file access, return open fd if possible */
int
vaccess(char *filename, int mode, int *err, long *sz)
{
	int		fd;
	char		*p, *q, fn[1024];
	struct stat	stbuf;

#if defined(WNAPP) || defined(NTAPP) || defined(NTSVC) || defined(NW) || defined(OS2)
	/* translate / -> \ */
	q= filename;
	while ((q= strchr (q, '/')) != 0)
		*q++= '\\';
#endif

	/* create filename with restricted access */
	if (TFTPDRestrict) {
		/* strip off absolute addressing and volume name */
		p= TFTPDChdir;
		if (!strrchr (filename, ':'))	/* has no volume name */
			if ((q= strrchr (p, ':')) != 0)
				p= q+1;
		
		/* check if already within the path */
		if (strncmp (p, filename, strlen (p)) == 0) {
			sprintf (fn, "%s", filename);
			log2("TFTPD: filename %s passed restriction\n",
				filename);
		}
		else {
			/* strip off absolute addressing and volume name */
			p= filename;
			if ((q= strrchr (p, ':')) != 0)	/* remove volume name */
				p= q+1;
			while (*p && !isalnum (*p))	/* first alpha */
				p++;
			if (*p == 0)			/* nothing left */
				{ *err= EACCESS; return (0); }
			
#ifdef UNIX
			sprintf (fn, "%s/%s", TFTPDChdir, p);
#else
			sprintf (fn, "%s\\%s", TFTPDChdir, p);
#endif
			log2("TFTPD: filename %s restricted to %s\n",
				filename, fn);
		}
	}
	else
		sprintf (fn, "%s", filename);

	/* check if present, return TFTP error code if failed */
	*err= 0;
	if (stat (fn, &stbuf) < 0 && mode == P_RRQ) {
		*err= errno == ENOENT ? ENOTFOUND : EACCESS;
		return (-1);
	}

	/* check if read/write able, returns TFTP error code */
	if (mode == P_RRQ) {
		if ((stbuf.st_mode & S_IROTH) == 0)
			{ *err= EACCESS; return (-1); }
	}
	else {
		if (!TFTPDCreat && (stbuf.st_mode & S_IWOTH) == 0)
			{ *err= EACCESS; return (-1); }
	}

	/* open the file */
	if (mode == P_RRQ) {
		if ((fd= mopen (fn, P_RRQ)) < 0)
			{ *err= errno + 100; return (-1); }
	}
	else {
		if ((fd= mopen (fn, P_WRQ)) < 0)
			{ *err= errno + 100; return (-1); }
	}

	*sz= stbuf.st_size;
	return (fd);
}


/* TDEL - delete a connection */
void
tdel(int t)
{
	/* now count this as a TFTP transfer */
	Trans[t].done= 0;
	Trans[t].first= 0;
	if (Trans[t].fd >= 0) {
		mclose (Trans[t].fd);
		if (Trans[t].abuf) 
			free(Trans[t].abuf);
		Trans[t].abuf= 0;
		Trans[t].fd= -1;
	}
	if (Trans[t].fname) 
		free(Trans[t].fname);
	Trans[t].fname = (char *) 0;
}


/* TPTR - return an index to the connection only if a new one
is created for a multicast file */
int
mptr(char *fname, SI *faddr)
{
	int	i;
	time_t	tm;

	/* don't leave any structure with the client addr and port #
	matching but for a different file name in the array
	this would confuse tptr to identify the transaction structure
	for an ack pkt */

	/* kill all previous connectionfrom this client.
	one client can have one active session. It is requesting for one now,
	so, it should not have any thing already */

	for (i= 0; i < TFTPDMaxTrans; i++) 
	{
		if (Trans[i].first != 0 && 
			Trans[i].si.sin_addr.s_addr == faddr->sin_addr.s_addr &&
			Trans[i].si.sin_port == faddr->sin_port) 
		{
			/* one client can have only one transfer at a time.
			if this is done delete it otherwise error */
			tdel(i);
		}
	}

	/* now see if anybody else is transferring this file,
	if so, don't let this client start it */

	for (i= 0; i < TFTPDMaxTrans; i++) 
	{
		if (Trans[i].first != 0 && Trans[i].msi.sin_addr.s_addr != 0 &&
					Trans[i].fname != (char *)0 && !Trans[i].done)
		{
			if (stricmp(fname, Trans[i].fname) == 0)
				return -1;	/* no need to process this open, file already
							being transferred */
		}
	}
	/* create a new connection for this client */

retry:
	/* not found, make up a new connection */
	for (i= 0; i < TFTPDMaxTrans; i++) {
		if (Trans[i].first == 0) {
			Trans[i].si= *faddr;
			time (&Trans[i].first);		/* tag as used */
			Trans[i].last= Trans[i].first;
			Trans[i].retry= 0;
			Trans[i].lpack= 0;
			Trans[i].msi.sin_addr.s_addr= 0;
			Trans[i].ucnt= 0;
			Trans[i].fd= -1;
			Trans[i].fsz= 0;
			Trans[i].done= 0;
			Trans[i].abuf= 0;
			Trans[i].alen= 0;
			Trans[i].segsize = DEFAULT_SEG_SIZE;
			/* copy the relative path name in to the structure */
			if ((Trans[i].fname = malloc(strlen(fname))) != 0)
				strcpy(Trans[i].fname, fname);
			return (i);
		}
	}

	/* do garbage collection */
	time (&tm);
	tm-= TFTPDTmoutval;
	for (i= 0; i < TFTPDMaxTrans; i++) 
	{
		if (Trans[i].first != 0 && Trans[i].last <= tm) {
			if (Trans[i].fd >= 0) {
				mclose (Trans[i].fd);
				if (Trans[i].abuf)
					free(Trans[i].abuf);
				Trans[i].abuf= 0;
				Trans[i].fd= -1;

				if (Trans[i].fname) 
					free(Trans[i].fname);
				Trans[i].fname = (char *) 0;
			}
			log2("TFTPD[%02d]: recycled "
				"%sused transmission structure\n",
				i, Trans[i].done ? "un" : "");

			Trans[i].first= 0;
			Trans[i].done= 0;
			goto retry;
		}
	}

	return (-2);
}


/* TPTR - return an index to the connection (negative if new) */
int
tptr(SI *faddr)
{
	int	i;
	time_t	tm;

	/* search the connection */
	for (i= 0; i < TFTPDMaxTrans; i++) {
		if (Trans[i].first != 0 && 
			Trans[i].si.sin_addr.s_addr == faddr->sin_addr.s_addr &&
			Trans[i].si.sin_port == faddr->sin_port) {
			if (!Trans[i].done)
				time (&Trans[i].last);	/* mark used */
			return (i+1);
		}
	}

retry:
	/* not found, make up a new connection */
	for (i= 0; i < TFTPDMaxTrans; i++) {
		if (Trans[i].first == 0) {
			Trans[i].si= *faddr;
			time (&Trans[i].first);		/* tag as used */
			Trans[i].last= Trans[i].first;
			Trans[i].retry= 0;
			Trans[i].lpack= 0;
			Trans[i].msi.sin_addr.s_addr= 0;
			Trans[i].ucnt= 0;
			Trans[i].fd= -1;
			Trans[i].fsz= 0;
			Trans[i].done= 0;
			Trans[i].abuf= 0;
			Trans[i].alen= 0;
			Trans[i].segsize = DEFAULT_SEG_SIZE;
			return (-(i+1));
		}
	}

	/* do garbage collection */
	time (&tm);
	tm-= TFTPDTmoutval;
	for (i= 0; i < TFTPDMaxTrans; i++) {
		if (Trans[i].first != 0 && Trans[i].last <= tm) {
			if (Trans[i].fd >= 0) {
				mclose (Trans[i].fd);
				if (Trans[i].abuf)
					free(Trans[i].abuf);
				Trans[i].abuf= 0;
				Trans[i].fd= -1;
				if (Trans[i].fname) 
					free(Trans[i].fname);
				Trans[i].fname = (char *) 0;
			}
			log2("TFTPD[%02d]: recycled "
				"%sused transmission structure\n",
				i, Trans[i].done ? "un" : "");

			Trans[i].first= 0;
			Trans[i].done= 0;
			goto retry;
		}
	}

	return (0);
}


/* TCLEAN - do a garbage collection */
void
tclean()
{
	int	i;
	time_t	tm;

	time (&tm);
	tm-= TFTPDTmoutval;
	for (i= 0; i < TFTPDMaxTrans; i++) {
		if (Trans[i].first != 0 && Trans[i].last <= tm) {
			if (Trans[i].fd >= 0) {
				mclose (Trans[i].fd);
				if (Trans[i].abuf)
					free(Trans[i].abuf);
				Trans[i].abuf= 0;
				Trans[i].fd= -1;
			}
			log2("TFTPD[%02d]: cleared "
				"%sused transmission structure\n",
				i, Trans[i].done ? "un" : "");

			Trans[i].first= 0;
			Trans[i].done= 0;
			if (Trans[i].fname) 
				free(Trans[i].fname);
			Trans[i].fname = (char *) 0;
		}
	}

	return;
}


/* OPENREAD - open a TFTP transmission for reading */
void 
openread(int s, int t, char *buf, int blen)
{
	TH	*dp, *tp;
	char	*cp, *mode, *filename;
	int	first, r, err, segsz;


#ifdef _DEBUG
	sprintf(FileName, "Sock=%d, index=%d, s_addr_s=%lx, s_addr_m=%lx\r\n", 
			s, t, Trans[t].si.sin_addr.s_addr, Trans[t].msi.sin_addr.s_addr);
	DEBUGMSG(_TEXT("%s"), FileName);
#endif
	/* extract the filename from the initial packet */
	tp= (TH *)buf;
	first= 1;
	cp = tp->th_stuff;

#ifdef _DEBUG
//	DEBUG_CONSTS(_TEXT("in openread\r\n"));
#endif
	if (tp->th_stuff[1] == ':' || tp->th_stuff[0] == '\\' ||
		 tp->th_stuff[0] == '/')	/* full path given */
	{
		filename = tp->th_stuff;
#ifdef _DEBUG
//	DEBUG_CONSTS(_TEXT("Full path given...\r\n"));
//	DEBUGMSG(_TEXT("path: %s.\r\n"), filename);
#endif
	}
	else
	{
		strcpy(FileName, Base_Path);
		strcat(FileName,tp->th_stuff);
		filename = FileName;
#ifdef _DEBUG
//	DEBUGMSG(_TEXT("Full path %s.\r\n"), filename);
#endif
	}

	if (s == LTFTPDSocket)
		segsz= LTFTPDSegsize;
	else
		segsz= Trans[t].segsize;

again:
	while (cp < buf + blen) {	/* while not end of package */
		if (!*cp)
			break;
		cp++;
	}

	if (*cp != 0) { 			/* must be 0-terminated */
		nak(s, &Trans[t].si, sizeof (SA), EBADOP);
		tdel(t);
		return;
	}

	/* now get the mode */
	if (first != 0) {
		mode= ++cp;
		first= 0;
		goto again;
	}

	/* convert mode to lowercase */
	for (cp= mode; *cp; cp++)
		if (isupper (*cp))
			*cp= tolower (*cp);

	/* call the requested function */
	if (strcmp (mode, MODE_NETASCII) != 0 && 
		strcmp (mode, MODE_OCTET) != 0) {
		nak(s, &Trans[t].si, sizeof (SA), EBADOP);
		tdel(t);
		return;
	}

	if ((Trans[t].fd= vaccess(filename, P_RRQ, &err, &Trans[t].fsz)) < 0) {
		log2("TFTPD[%02d vaccess/RRQ]: filename= %s, errno= %d,"
			" TFTPerror= %d\n", t, filename, errno, err);
		nak(s, &Trans[t].si, sizeof (SA), err);
		tdel(t);
		return;
	}

	if (TFTPDAhead)
		Trans[t].abuf= malloc (MSEGSIZE);

	/* start the file transfer */
	dp= (TH *)Buf;
	dp->th_opcode= htons(P_DATA);
	Trans[t].lpack= 1;
	dp->th_block= htons((unsigned short)Trans[t].lpack);
    	lseek (Trans[t].fd, (long)(Trans[t].lpack-1)*segsz, SEEK_SET);
	if ((r= read (Trans[t].fd, dp->th_data, segsz)) < 0) {
		log1("TFTPD[%02d read(1)]: TFTP block= 0x%04x, errno= %d,"
			" fd= %d\n", t, ntohs(dp->th_block),
			errno, Trans[t].fd);
		nak(s, &Trans[t].si, sizeof (SA), errno+100);
		tdel(t);
		return;
	}

	/* ...and send it out */
	if (Trans[t].msi.sin_addr.s_addr == 0 || Trans[t].ucnt > 0) {
		Trans[t].ucnt--;
		if (sendto (s, (char *)dp, HDRSIZE+r,
			0, (SA *)&Trans[t].si, sizeof (SA)) != HDRSIZE+r) {
			log1("TFTPD[%02d sendto(1)]: TFTP block= 0x%04x,"
				" errno= %d\n", t, ntohs(dp->th_block), errno);
			nak(s, &Trans[t].si, sizeof (SA), errno+100);
			tdel(t);
			return;
		}
	}

	if (Trans[t].msi.sin_addr.s_addr != 0) {
		if (sendto (s, (char *)dp, HDRSIZE+r,
			0, (SA *)&Trans[t].msi, sizeof (SA)) != HDRSIZE+r) {
			log1("TFTPD[%02d sendto(1m)]: TFTP block= 0x%04x,"
				" errno= %d\n", t, ntohs(dp->th_block), errno);
			nak(s, &Trans[t].si, sizeof (SA), errno+100);
			tdel(t);
			return;
		}
	}
}


/* OPENWRITE - open a TFTP transmission for writing */
void 
openwrite(int s, int t, char *buf, int blen)
{
	TH	*tp;
	char	*cp, *mode, *filename;
	int	first, err, segsz;

	
	/* extract the filename from the initial packet */
	tp= (TH *)buf;
	first= 1;
	filename= cp= tp->th_stuff;
	segsz= SEGSIZE;

again:
	while (cp < buf + blen) {	/* while not end of package */
		if (!*cp)
			break;
		cp++;
	}

	if (*cp != 0) { 			/* must be 0-terminated */
		nak(s, &Trans[t].si, sizeof (SA), EBADOP);
		tdel(t);
		return;
	}

	/* now get the mode */
	if (first != 0) {
		mode= ++cp;
		first= 0;
		goto again;
	}

	/* convert mode to lowercase */
	for (cp= mode; *cp; cp++)
		if (isupper (*cp))
			*cp= tolower (*cp);

	/* call the requested function */
	if (strcmp (mode, MODE_NETASCII) != 0 && 
		strcmp (mode, MODE_OCTET) != 0) {
		nak(s, &Trans[t].si, sizeof (SA), EBADOP);
		tdel(t);
		return;
	}

	if ((Trans[t].fd= vaccess(filename, P_WRQ, &err, &Trans[t].fsz)) < 0) {
		log2("TFTPD[%02d vaccess/WRQ]: filename= %s, errno= %d,"
			" TFTPerror= %d\n", t, filename, errno, err);
		nak(s, &Trans[t].si, sizeof (SA), err);
		tdel(t);
		return;
	}

	/* acknowledge receivement of first block */
	Trans[t].lpack= 0;
	Trans[t].fsz= 0;
	ack(s, &Trans[t].si, sizeof (SA), Trans[t].lpack);
}


/* RCVBLK - receive a TFTP block from client */
void
rcvblk(int s, int t, char *buf, int blen)
{
	TH	*tp;
	int	r;
	time_t	tm;
	int	segsz;
	
	segsz= SEGSIZE;
	tp= (TH *)buf;

	/* check for short package here */
	if (blen < HDRSIZE) {
		log1("TFTPD[%02d]: short package received\n", t);
		return; 
	}

	/* check for right order */
	if (ntohs(tp->th_block) != Trans[t].lpack+1) {
		Trans[t].retry++;

		/* synchronize with the other side */
		log2("TFTPD[%02d]: out of sync, srv= 0x%04x, clnt"
			"= 0x%04x, retry= %d\n", t,
			Trans[t].lpack,
			ntohs(tp->th_block), Trans[t].retry);

		/* acknowledge last received package */
		ack(s, &Trans[t].si, sizeof (SA), Trans[t].lpack);
		return;
	}

	/* count the received bytes */
	Trans[t].fsz+= blen-HDRSIZE;

	/* write block */
	if ((r= write (Trans[t].fd, tp->th_data,
		(blen-HDRSIZE))) != (blen-HDRSIZE)) {
		log1("TFTPD[%02d write(1)]: TFTP block= 0x%04x,"
			" errno= %d, fd= %d\n", t, Trans[t].lpack,
			errno, Trans[t].fd);
		nak(s, &Trans[t].si, sizeof (SA), errno+100);
		tdel(t); 
		return;
	}

	/* last package, print statistics */
	if ((blen-HDRSIZE) != segsz && Trans[t].done == 0) {
		Trans[t].done= 1;
		time (&tm);
		if (tm-Trans[t].first > 0)
			log2("TFTPD[%02d]: transfer= %d KByte/sec,"
				" bytes= %ld, retries= %d\n",
				t, Trans[t].fsz/(tm-Trans[t].first)/1024,
				Trans[t].fsz, Trans[t].retry);
		else
			log2("TFTPD[%02d]: transfer= %d KByte/sec,"
				" bytes= %ld, retries= %d\n", 
				t, Trans[t].fsz/1024,
				Trans[t].fsz, Trans[t].retry);
	
		/* close filedes */
		mclose (Trans[t].fd);
		Trans[t].fd= -1;
	}

	/* acknowledge block */
	Trans[t].lpack= ntohs(tp->th_block);
	ack(s, &Trans[t].si, sizeof (SA), ntohs(tp->th_block));

#ifdef FOO
	/* if client want's next block but we are finished just return */
	if (Trans[t].done) {
		log2("TFTPD[%02d]: reached end of file, fd= %d\n",
			t, Trans[t].fd);
		return;
	}

	/* send next block */
	dp= (TH *)buf;
	dp->th_opcode= htons(P_DATA);
	Trans[t].lpack= ntohs(tp->th_block)+1;
	dp->th_block= htons(Trans[t].lpack);

	if (Trans[t].abuf != 0 && Trans[t].alen >= 0
		&& Trans[t].apack == Trans[t].lpack) {
		tahead= 1;
		memcpy (dp->th_data, Trans[t].abuf, r= Trans[t].alen);
	}
	else {
		tahead= 0;
		lseek (Trans[t].fd,
			(long)(Trans[t].lpack-1)*segsz, SEEK_SET);
		if ((r= read (Trans[t].fd, dp->th_data, segsz)) < 0) {
			log1("TFTPD[%02d read(2)]: TFTP block= 0x%04x,"
				" errno= %d, fd= %d\n", t, Trans[t].lpack,
				errno, Trans[t].fd);
			nak(s, &Trans[t].si, sizeof (SA), errno+100);
			tdel(t); 
			return;
		}
	}
	
	/* last package, print statistics */
	if (r != segsz && Trans[t].done == 0) {
		Trans[t].done= 1;
		time (&tm);
		if (tm-Trans[t].first > 0)
			log2("TFTPD[%02d]: transfer= %d KByte/sec, retries= %d\n",
				t, Trans[t].fsz/(tm-Trans[t].first)/1024),
				Trans[t].retry);
		else
			log2("TFTPD[%02d]: transfer= %d KByte/sec, retries= %d\n",
				t, Trans[t].fsz/1024),
				Trans[t].retry);
	}


	/* ...and send it out */
	if (Trans[t].msi.sin_addr.s_addr == 0 || Trans[t].ucnt > 0) {
		Trans[t].ucnt--;
		if (sendto (s, (char *)dp, HDRSIZE+r,
			0, (SA *)&Trans[t].si, sizeof (SA)) != HDRSIZE+r) {
			log1("TFTPD[%02d sendto(3)]: TFTP block= 0x%04x,"
				" errno= %d\n", t, Trans[t].lpack, errno);
			nak(s, &Trans[t].si, sizeof (SA), errno+100);
			tdel(t);
			return;
		}
	}
	if (Trans[t].msi.sin_addr.s_addr != 0) {
		if (sendto (s, (char *)dp, HDRSIZE+r,
			0, (SA *)&Trans[t].msi, sizeof (SA)) != HDRSIZE+r) {
			log1("TFTPD[%02d sendto(3m)]: TFTP block= 0x%04x,"
				" errno= %d\n", t, Trans[t].lpack, errno);
			nak(s, &Trans[t].si, sizeof (SA), errno+100);
			tdel(t);
			return;
		}
	}

	if (Trans[t].abuf != 0) {
		if (tahead)
			lseek (Trans[t].fd, 
				(long)Trans[t].lpack*segsz, SEEK_SET);

		Trans[t].alen= read (Trans[t].fd, Trans[t].abuf, segsz);
		Trans[t].apack= Trans[t].lpack+1;
	}
#endif
}


/* SENDBLK - send a TFTP block to client */
void
sendblk(int s, int t, char *buf, int blen)
{
	TH	*tp, *dp;
	int	r;
	time_t	tm;
	int	segsz, tahead;
	
	if (s == LTFTPDSocket)
		segsz= LTFTPDSegsize;
	else
#ifdef AVL
		segsz= SEGSIZE;
#endif
		segsz= Trans[t].segsize;

	/* check for short package here */
	if (blen < HDRSIZE) {
		log1("TFTPD[%02d]: short package received\n", t);
		return; 
	}

	tp= (TH *)buf;
	if (ntohs(tp->th_block) != Trans[t].lpack) {
		Trans[t].retry++;

		/* synchronize with the other side */
		log2("TFTPD[%02d]: out of sync, srv= 0x%04x, clnt"
			"= 0x%04x, retry= %d\n", t,
			Trans[t].lpack,
			ntohs(tp->th_block), Trans[t].retry);
	}

	/* if client want's next block but we are finished just return */
	/* AVL - check if the request is for last packet! it might have been
		lost and we may have to resend */
	if (Trans[t].done) 
	{	/* after sending the last pkt, accept requests only for the last one */
		if (ntohs(tp->th_block) != Trans[t].lpack -1) 
		{	/* not an ack for last-1,  ignore it */
			log2("TFTPD[%02d]: reached end of file, fd= %d\n",
				t, Trans[t].fd);
			return;
		}
		else
			Trans[t].done = 0;	/* not done yet! */
	}

	/* send next block */
	dp= (TH *)buf;
	dp->th_opcode= htons(P_DATA);
	Trans[t].lpack= ntohs(tp->th_block)+1;
	dp->th_block= htons((unsigned short)Trans[t].lpack);

	if (Trans[t].abuf != 0 && Trans[t].alen >= 0
		&& Trans[t].apack == Trans[t].lpack) {
		tahead= 1;
		memcpy (dp->th_data, Trans[t].abuf, r= Trans[t].alen);
	}
	else {
		tahead= 0;
		lseek (Trans[t].fd,
			(long)(Trans[t].lpack-1)*segsz, SEEK_SET);
		if ((r= read (Trans[t].fd, dp->th_data, segsz)) < 0) {
			log1("TFTPD[%02d read(2)]: TFTP block= 0x%04x,"
				" errno= %d, fd= %d\n", t, Trans[t].lpack,
				errno, Trans[t].fd);
			nak(s, &Trans[t].si, sizeof (SA), errno+100);
			tdel(t); 
			return;
		}
	}
	
	/* last package, print statistics */
	if (r != segsz && Trans[t].done == 0) {

#ifdef _DEBUG
	DEBUGMSG(_TEXT("Last pkt on %d\r\n"), t);
#endif
		Trans[t].done= 1;
		time (&tm);
		if (tm-Trans[t].first > 0)
			log2("TFTPD[%02d]: transfer= %d KByte/sec, len= %ld,"
				" retries= %d\n",
				t, Trans[t].fsz/(tm-Trans[t].first)/1024,
				Trans[t].fsz, Trans[t].retry);
		else
			log2("TFTPD[%02d]: transfer= %d KByte/sec, len= %d,"
				" retries= %d\n",
				t, Trans[t].fsz/1024,
				Trans[t].fsz, Trans[t].retry);
	}

	/* ...and send it out */
	if (Trans[t].msi.sin_addr.s_addr == 0 || Trans[t].ucnt > 0) {
		Trans[t].ucnt--;
		if (sendto (s, (char *)dp, HDRSIZE+r,
			0, (SA *)&Trans[t].si, sizeof (SA)) != HDRSIZE+r) {
			log1("TFTPD[%02d sendto(3)]: TFTP block= 0x%04x,"
				" errno= %d\n", t, Trans[t].lpack, errno);
			nak(s, &Trans[t].si, sizeof (SA), errno+100);
			tdel(t);
			return;
		}
	}
	if (Trans[t].msi.sin_addr.s_addr != 0) {
		if (sendto (s, (char *)dp, HDRSIZE+r,
			0, (SA *)&Trans[t].msi, sizeof (SA)) != HDRSIZE+r) {
			log1("TFTPD[%02d sendto(3m)]: TFTP block= 0x%04x,"
				" errno= %d\n", t, Trans[t].lpack, errno);
			nak(s, &Trans[t].si, sizeof (SA), errno+100);
			tdel(t);
			return;
		}
	}

	if (Trans[t].abuf != 0) {
		if (tahead)
			lseek (Trans[t].fd, 
				(long)Trans[t].lpack*segsz, SEEK_SET);

		Trans[t].alen= read (Trans[t].fd, Trans[t].abuf, segsz);
		Trans[t].apack= Trans[t].lpack+1;
	}
}


/* HANDLE - handles new and existing connections */
void
handle(int s, char *req)
{
	int	blen, flen, i, o, t;
	TH	*tp;
	SI	faddr;
	char TmpBuf[32], *q;
	LONG ret_val;
	DWORD ValType;
	DWORD ValSize, MTFTPAddr;
	unsigned char d1,d2,d3,d4;

	flen= sizeof (faddr);
	if ((blen= recvfrom (s, Buf, sizeof (Buf), 0,
		(SA *)&faddr, &flen)) <= 0) 
	{
#ifdef NTSVC
		if(StopErrno == FALSE)
		{
#endif
			log1("TFTPD[recvfrm()]: errno= %d\n", errno);
#ifdef NTSVC
		}
#endif
		return;
	}
 
reset:
	tp= (TH *)Buf;

	/* create a new connection only in case of unicast. Multicast
	should not start twice for the same file */

	/* identifying the transaction for a mcast open request
	is different from mcast acks or any unicast packet. 
	in the former case, we need to check the file names and in the 
	later case we can check with the clients address */

	if (s == MTFTPDSocket && ntohs(tp->th_opcode) == P_RRQ) /* multicast open */
	{
#ifdef _DEBUG
	DEBUGMSG(_TEXT("mcst open req:  %s.\r\n"), tp->th_stuff);
#endif

		if ((o= mptr(tp->th_stuff, &faddr)) == -1) 
		{
#ifdef _DEBUG
	DEBUGMSG(_TEXT("2nd open req for %s.\r\n"), tp->th_stuff);
#endif
			/* don't start now, already going */
			return;
		}
		else if (o == -2)
		{
			log1("TFTPD: TFTP transfers exhausted,"
				" rejecting until running TFTP transfer expires\n");
			nak(s, &faddr, sizeof (SA), EBADOP);
			return; 
		}
		else t = o;	/* always new connection, never an old one at this point */
		o = -(o+1);	/* to satisfy the other checks on 'o' */
#ifdef _DEBUG
	DEBUGMSG(_TEXT("New MTFTP connection: %d.\r\n"), t);
#endif

	}

	else
	{
		/* unknown connection */
		if ((o= tptr(&faddr)) == 0) {
			log1("TFTPD: TFTP transfers exhausted,"
				" rejecting until running TFTP transfer expires\n");
			nak(s, &faddr, sizeof (SA), EBADOP);
			return; 
		}
		if (o < 0)
			t= -o-1; 	/* new connection */
		else
			t= o-1;		/* old connection */
	}

	/* check for short package here */
	if (blen < HDRSIZE) {
		log1("TFTPD[%02d]: short package received\n", t);
		return; 
	}

	/* process that package */
	switch (ntohs(tp->th_opcode)) {
		case P_RRQ:
			/* if old transmission reset now */
			if (o > 0) {
				log2("TFTPD[%02d]: reset transmission"
					" structure\n", t);
				tdel(t);
				goto reset;
			}
			
			log2("TFTPD[%02d]: %s, fn= %s, mode= %s,"
				" from= %d.%d.%d.%d/%u\n",
				t, req, tp->th_stuff,
				tp->th_stuff + strlen (tp->th_stuff) + 1,
				(short)(ntohl(Trans[t].si.sin_addr.s_addr)>>24)
				& 0xff,
				(short)(ntohl(Trans[t].si.sin_addr.s_addr)>>16)
				& 0xff,
				(short)(ntohl(Trans[t].si.sin_addr.s_addr)>> 8)
				& 0xff,
				(short)ntohl(Trans[t].si.sin_addr.s_addr)
				& 0xff,
				ntohs(Trans[t].si.sin_port));

			i= 0;
			if (s == MTFTPDSocket) 
			{
#ifdef AVL
			    while (Mtftpc[i].fnam[0] && i < MTFTPDMAXFILES) 
				{
					if (!strcmp (tp->th_stuff, Mtftpc[i].fnam)) 
					{
						Trans[t].msi.sin_addr.s_addr=
							htonl(Mtftpc[i].addr);
						Trans[t].msi.sin_family= AF_INET;
						Trans[t].msi.sin_port= htons((unsigned short)Mtftpc[i].port);
						Trans[t].ucnt= TFTPDUnicast;
						if (Trans[t].ucnt == 0)
							Trans[t].ucnt= 32767;
						break;
					}
					i++;
			    }
#else	/* AVL */ 
				{	
				/* translate / -> \ */
				q= tp->th_stuff;
				while ((q= strchr (q, '/')) != 0)
					*q++= '\\';
				}
				ValSize = 32;
				ret_val = RegQueryValueEx(hTftpdFilesKey, tp->th_stuff, NULL, 
						&ValType, (LPBYTE)TmpBuf, &ValSize);
				/* if this file name is not found in the registry,
				send an error pkt so that client will try TFTP
				it doesn't have to time out on MTFTP */

				if (ret_val || sscanf(TmpBuf, "%d.%d.%d.%d", &d1, &d2, &d3, &d4) < 0)
				{
					nak(s, &Trans[t].si, sizeof (SA), EMCASTADDR);
					tdel(t);
					return;
				}
				else
				{ 
					MTFTPAddr= ((long)d1 << 24) + ((long)d2 << 16) +
									(d3 << 8) + d4;
					Trans[t].msi.sin_addr.s_addr=htonl(MTFTPAddr);
					Trans[t].msi.sin_family= AF_INET;
					Trans[t].msi.sin_port= htons((unsigned short)MTFTPPort);
					Trans[t].ucnt= TFTPDUnicast;
				}
#endif
			    if (Trans[t].msi.sin_addr.s_addr == 0) 
				{
					log1("TFTPD[%02d]: %s is not in %s,"
					" using TFTP\n", t,
					tp->th_stuff,
					TFTPDMconfig);
			    }
				Trans[t].segsize = McastSegSize; 
			}
			else	/* tftpd connection, process options */
			{
				DWORD FileSize;
				int BlockSize;
				int i = 0;
				struct stat	stbuf;

				FileSize = 0;
				BlockSize = 0;
				/* position i where the options start in th_stuff */
				/* skip file name */

				i += (strlen(&tp->th_stuff[i]) + 1);
				/* skip mode string */
				i += (strlen(&tp->th_stuff[i]) + 1);
				
				while (i < blen-2)	/* till the end of pkt */
				{
					if (!stricmp(&tp->th_stuff[i], "blksize"))
					{
						/* get the block size */
						i += (strlen(&tp->th_stuff[i]) + 1);
						BlockSize = atoi (&tp->th_stuff[i]);
						if (BlockSize < DEFAULT_SEG_SIZE)
							BlockSize = 0;
						else if (BlockSize > MAX_DATA_SIZE)	
							BlockSize = MAX_DATA_SIZE;
						/* compare it with server max and send an oack */
						/* in any case skip this value field */
						i += (strlen(&tp->th_stuff[i]) + 1);
					}
					else if (!stricmp(&tp->th_stuff[i], "tsize"))
					{
						/* skip the option field */
						i += (strlen(&tp->th_stuff[i]) + 1);
						/* send the file size in the oack*/

						/* translate / -> \ in filename */
						q= tp->th_stuff;
						while ((q= strchr (q, '/')) != 0)
						*q++= '\\';
						if (tp->th_stuff[1] != ':' && tp->th_stuff[0] != '\\')
						{
							strcpy(FileName, Base_Path);
							strcat (FileName, tp->th_stuff);
						}
						else strcpy (FileName, tp->th_stuff);
						if (stat (FileName, &stbuf) == 0)
							FileSize = stbuf.st_size;
						/* skip the value part of it */
						i += (strlen(&tp->th_stuff[i]) + 1);
					}
					else
					{
						/* skip the unknown option & it's value string */
						i += (strlen(&tp->th_stuff[i]) + 1);
						i += (strlen(&tp->th_stuff[i]) + 1);
					}
				}
				if (BlockSize || FileSize)	/* send oack */
				{
					if (BlockSize > DEFAULT_SEG_SIZE)
						Trans[t].segsize = BlockSize;
					oack(s, &faddr, sizeof (SA), FileSize, BlockSize);
				}
			}
			
			openread(s, t, Buf, blen);
			return;
		
		case P_WRQ:
			/* if old transmission reset now */
			if (o > 0) {
				log2("TFTPD[%02d]: reset transmission"
					" structure\n", t);
				tdel(t);
				goto reset;
			}

			log1("TFTPD[%02d]: WRQ, fn= %s, mode= %s,"
				" from= %d.%d.%d.%d/%u\n",
				t, tp->th_stuff,
				tp->th_stuff + strlen (tp->th_stuff) + 1,
				(short)(ntohl(Trans[t].si.sin_addr.s_addr)>>24)
				& 0xff,
				(short)(ntohl(Trans[t].si.sin_addr.s_addr)>>16)
				& 0xff,
				(short)(ntohl(Trans[t].si.sin_addr.s_addr)>> 8)
				& 0xff,
				(short)ntohl(Trans[t].si.sin_addr.s_addr)
				& 0xff,
				ntohs(Trans[t].si.sin_port));

			/* are TFTP writes enabled */
			if (!TFTPDWrite) {
				log1("TFTPD[%02d]: TFTP writes are not"
					" enabled (-w option) !\n", t);
				nak(s, &Trans[t].si, sizeof (SA), EBADOP);
				tdel(t);
			}

			openwrite(s, t, Buf, blen);
			return;

		case P_ACK:
#ifdef _DEBUG
	DEBUGMSG(_TEXT("ACK: %d. "), t);
#endif

			if (o < 0) {
				log2("TFTPD[%02d]: suspicious package\n", t);
				tdel(t);
				return;
			}
			sendblk(s, t, Buf, blen);
			return;

		case P_DATA:
			if (o < 0) {
				log2("TFTPD[%02d]: suspicious package\n", t);
				tdel(t);
				return;
			}
			rcvblk(s, t, Buf, blen);
			return;

		default:
			log1("TFTPD: unsupported opcode= 0x%04x,"
				" from= %d.%d.%d.%d/%u\n",
				ntohs(tp->th_opcode),
				(short)(ntohl(Trans[t].si.sin_addr.s_addr)>>24)
				& 0xff,
				(short)(ntohl(Trans[t].si.sin_addr.s_addr)>>16)
				& 0xff,
				(short)(ntohl(Trans[t].si.sin_addr.s_addr)>> 8)
				& 0xff,
				(short)ntohl(Trans[t].si.sin_addr.s_addr)
				& 0xff,
				ntohs(Trans[t].si.sin_port));
			nak(s, &Trans[t].si, sizeof (SA), EBADOP);
			tdel(t);
			return;
	}
}


#if defined(WNAPP) || defined(NTAPP) || defined(NTSVC)
int
Mmain(int argc, char **argv)
#else
int
main(int argc, char **argv)
#endif
{
#ifdef WNAPP
	struct servent far	*sp;
#else
	struct servent		*sp;
#endif
#ifdef AVL
	FILE			*fp;
#endif	/* AVL */
	SI			saddr;
	char			*q;
#ifdef WNAPP
	char		about[128], title[64];
#endif
	int			rc, i, n, l;
#ifdef AVL
	int			d1, d2, d3, d4;
#endif	/* AVL */
	long		tcnt;
	fd_set			rfd;
	struct timeval		tmout;
DWORD avl_ret;
char TmpBuf[256];
DWORD ValType;
DWORD ValSize, ValSize1, ret_val;

#ifdef NW
	NETDB_DEFINE_CONTEXT;
#endif

#ifdef NTAPP
	/* prevent warnings */
	tmout= tmout;
	rfd= rfd;
	l= l;
#endif

#ifdef WNAPP
	doclose= 0;
	sprintf (about, "%s Version %s (%d Bit)\n%s",
		DESCRIPTION, TFTPD_VERSION, 16, COPYRIGHT);
	sprintf (title, "%s %s", DESCRIPTION, TFTPD_VERSION);
	winio_about(about);
	winio_settitle(winio_current(), title);
	{
		WORD x;
		WSADATA y;

		x= MAKEWORD(2,0);
		WSAStartup(x, &y);
	}
	winio_onclose(__hMainWnd, (DESTROY_FUNC)TFTPDatexit);
#endif

	/* initialize global data */
	TFTPDRestrict= DEF_TFTPDRESTRICT;
	TFTPDVerbose= DEF_TFTPDVERBOSE;
	TFTPDMaxTrans= DEF_TFTPDMAXTRANS;
	TFTPDTmoutval= DEF_TFTPDTMOUTVAL;
	TFTPDUnicast= DEF_TFTPDUNICAST;
	TFTPDAhead= DEF_TFTPDAHEAD;
	TFTPDMconfig= 0;
	TFTPDChdir= 0;
	TFTPDLogFil= 0;
	TFTPDLogFp= 0;
	LTFTPDSocket= -1;
	MTFTPDSocket= -1;
	LTFTPDFlag= 0;
	MTFTPDFlag= 0;
	TFTPDCreat= 0;
	TFTPDWrite= 0;

	n= 1;
	tcnt= 0;
	while (n < argc) {
	    if (argv[n][0] == '-')
	    	switch (argv[n][1]) {
			case 'c':
			case 'C':
				if (!argv[n+1])	
					usage();
				n++;
				if (argv[n][0] < '0' || argv[n][0] > '9')
					usage();
				i= atoi (argv[n]);
				if (i <= 0 || i > TFTPDMAXTRANS)
					usage();
				TFTPDMaxTrans= i;
				n++;
				break;

			case 'd':
			case 'D':
				if (argv[n+1] == 0 || argv[n+1][0] == 0 ||
					argv[n+1][0] == '-')
					TFTPDChdir= DEF_TFTPDCHDIR;
				else {
					TFTPDChdir= argv[n+1];
					n++;
				}
				n++;
				break;

			case 'h':
			case 'H':
				TFTPDAhead++;
				n++;
				break;

			case 'i':
			case 'I':
				if (!argv[n+1])	
					usage();
				n++;
				if (argv[n][0] < '0' || argv[n][0] > '9')
					usage();
				i= atoi (argv[n]);
				if (i < 0)
					usage();
				TFTPDTimeout= i;
				n++;
				break;

			case 'k':
			case 'K':
				if (!argv[n+1])	
					usage();
				n++;
				if (argv[n][0] < '0' || argv[n][0] > '9')
					usage();
				i= atoi (argv[n]);
				if (i < 0)
					usage();
				TFTPDTmoutval= i;
				n++;
				break;

			case 'l':

			case 'L':
				if (argv[n+1] == 0 || argv[n+1][0] == 0 ||
					argv[n+1][0] == '-')
					TFTPDLogFil= DEF_TFTPDLOGFIL;
				else {
					TFTPDLogFil= argv[n+1];
					n++;
				}
				n++;
				break;

			case 'm':
			case 'M':
				n++;
				if (argv[n] == 0 || argv[n][0] == 0 ||
					argv[n][0] == '-')
					usage();
#ifdef AVL
				TFTPDMconfig= argv[n];
#endif	/* AVL */
				
				n++;
#ifdef AVL
				if (argv[n] == 0 || argv[n][0] == 0 ||
					argv[n][0] < '0' ||
					argv[n][0] > '9')
					usage();
				i= atoi (argv[n]);
				if (i <= 0)
					usage();
				MTFTPDPort= ntohs((unsigned short)i);
				MTFTPDFlag++;
#endif	/* AVL */
				n++;
				break;

			case 'p':
			case 'P':
				if (!argv[n+1])	
					usage();
				n++;
				if (argv[n][0] < '0' || argv[n][0] > '9')
					usage();
				i= atoi (argv[n]);
				if (i <= 0)
					usage();
				TFTPDPort= ntohs((unsigned short)i);
				n++;
				break;

			case 'r':
			case 'R':
				TFTPDRestrict++;
				n++;
				break;

			case 's':
			case 'S':
				n++;
				if (argv[n] == 0 || argv[n][0] == 0 ||
					argv[n][0] < '0' ||
					argv[n][0] > '9')
					usage();
				i= atoi (argv[n]);
				if (i <= 0 || i > MSEGSIZE)
					usage();
				LTFTPDSegsize= i;
				
				n++;
				if (argv[n] == 0 || argv[n][0] == 0 ||
					argv[n][0] < '0' ||
					argv[n][0] > '9')
					usage();
				i= atoi (argv[n]);
				if (i <= 0)
					usage();
				LTFTPDPort= ntohs((unsigned short)i);
				LTFTPDFlag++;

				n++;
				break;

			case 'u':
			case 'U':
				n++;
#ifdef AVL
				if (argv[n] == 0 || argv[n][0] == 0 ||
					argv[n][0] < '0' ||
					argv[n][0] > '9')
					usage();
				i= atoi (argv[n]);
				if (i < 0)
					usage();
				TFTPDUnicast= i;
#endif	/* AVL */
				n++;
				break;

			case 'v':
			case 'V':
				if (argv[n+1])
					TFTPDVerbose= argv[n+1][0]-'0';
				else
					usage();
				if (TFTPDVerbose < 0 || TFTPDVerbose > 2)
					usage();
				n+= 2;
				break;

			case 'w':
			case 'W':
				TFTPDWrite++;
				n++;
				break;

			case 'x':
			case 'X':
				TFTPDCreat++;
				n++;
				break;

			default:
				usage();
		}
	    else
	    	usage();
    	}

#ifdef NW
	/* close NLM screen if requested */
	if (TFTPDVerbose == 0)
		DestroyScreen (GetCurrentScreen ());
#endif

#ifdef AVL
	AVLFp = fopen ("d:\\tftpd.txt", "w");
#endif	/* AVL */

	/* open the MTFTP configuration file */
#ifdef AVL
	Mtftpc[0].fnam[0]= 0;
	if (TFTPDMconfig) {
#if defined(WNAPP) || defined(NTAPP) || defined(NTSVC) || defined(NW) || defined(OS2)
		/* translate / -> \ */ 
		q= TFTPDMconfig;
		while ((q= strchr (q, '/')) != 0)
			*q++= '\\';
#endif

		if ((fp= fopen (TFTPDMconfig, "r")) == 0) {
			log0("TFTPD[fopen()]: fn= %s, errno= %d\n",
				TFTPDMconfig, errno);
			exit (1);
		}

		i= 0;
		while (!feof (fp) && i < MTFTPDMAXFILES) {
			if (fscanf (fp, "%s %d.%d.%d.%d %d",
				Mtftpc[i].fnam, &d1, &d2, &d3, &d4,
				&Mtftpc[i].port) < 0)
				break;
#if defined(WNAPP) || defined(NTAPP) || defined(NTSVC) || defined(NW) || defined(OS2)
			/* translate / -> \ */ 
			q= Mtftpc[i].fnam;
			while ((q= strchr (q, '/')) != 0)
				*q++= '\\';
#endif
			Mtftpc[i].addr= ((long)d1 << 24) + ((long)d2 << 16) +
				(d3 << 8) + d4;
			i++;
		}
		Mtftpc[i].fnam[0]= 0;

		fclose (fp);
	}
#endif	/* AVL */

#ifdef _DEBUG
	AllocConsole();
	DEBUG_CONSTS(_TEXT("Starting to process MTFTPD keys.\r\n"));
#endif

#ifdef NTSVC
	/* open the mtftpd registry keys */
	if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, KEY_MTFTPD, 0, KEY_READ, &hTftpdKey))
	{
#ifdef _DEBUG
	DEBUG_CONSTS(_TEXT("Failed to open MTFTPD key.\r\n"));
#endif
	}
	else if (avl_ret = RegOpenKeyEx(hTftpdKey, MTFTPD_FILES_KEY, 0, 
			KEY_ALL_ACCESS, &hTftpdFilesKey))
	{
#ifdef _DEBUG
	DEBUGMSG(_TEXT("Failed to open Files key %d.\r\n"), avl_ret);
#endif
	}
	else	/* files key exists, give addresses to all files */
	{
		int j, use_unicast = 0;		
		unsigned char start_addr[4], address_buf[16];	/* a.b.c.d */

		ValSize = ValSize1 = 256;
		if (RegQueryValueEx(hTftpdKey, MCAST_ENABLE_VALUE, NULL, &ValType, 
										(LPBYTE)TmpBuf, &ValSize) ||
			(*(DWORD *)TmpBuf == 0L) || 
			RegQueryValueEx(hTftpdKey, MCAST_START_ADDRESS_VALUE, NULL, 
									&ValType, (LPBYTE)TmpBuf, &ValSize1) ||
			ValSize1 == 0)
		{
#ifdef _DEBUG
	DEBUG_CONSTS(_TEXT("enable or start_addr failed\r\n"));
#endif

			MTFTPDPort= 0;
			MTFTPPort= 0;
			MTFTPDFlag = 0;
			/* assign 0.0.0.0 for all files */
		}
		else	/* get the client and server ports */
		{
#ifdef _DEBUG
	DEBUG_CONSTS(_TEXT("start_addr read\r\n"));
#endif
			/* remember the start address */
			i = sscanf(TmpBuf, "%d.%d.%d.%d", &start_addr[0], &start_addr[1],
							&start_addr[2], &start_addr[3]);

			ValSize = ValSize1 = 256;
			/* check if start_addr is a valid multicast addr */
			if (i != 4 || start_addr[0] < 224 || start_addr[0] > 239 ||
				RegQueryValueEx(hTftpdKey, MCAST_SRVR_PORT_VALUE, NULL, 
							&ValType, (LPBYTE)TmpBuf, &ValSize) ||
				((i = atoi(TmpBuf)) < 0) || 	
				RegQueryValueEx(hTftpdKey, MCAST_CLNT_PORT_VALUE, NULL, 
							&ValType, (LPBYTE)TmpBuf, &ValSize1) ||
				((j = atoi(TmpBuf)) < 0))
			{
#ifdef _DEBUG
	DEBUG_CONSTS(_TEXT("One of the port values failed\r\n"));
#endif
				MTFTPDPort= 0;
				MTFTPPort= 0;
				MTFTPDFlag = 0;
			}
			else
			{	
				MTFTPDPort= ntohs((unsigned short)i);
				MTFTPPort= j;
				MTFTPDFlag++;
#ifdef _DEBUG
	DEBUGMSG(_TEXT("port values: %d"), MTFTPPort);
	DEBUGMSG(_TEXT(" %d\r\n"), MTFTPDPort);
#endif
			}
		}	/* got the port values */

		/* assign the addresses to the files */
		/* enumerate all values and assign data to them */
		ValSize1 = 0;	/* used as index for the value */
		ret_val = 0;
		if (!MTFTPDFlag)	/* not enabled or start address not known */
		{		/* fill the address buffer with 0.0.0.0 */
			i = sprintf(address_buf, "%d.%d.%d.%d", 0,0,0,0); 
			use_unicast = 1;
		}
		while(ret_val == 0)
		{
			ValSize = 256;
			ret_val = RegEnumValue(hTftpdFilesKey, ValSize1, (LPBYTE)TmpBuf, 
						&ValSize, NULL, &ValType, NULL, NULL);
			if (ret_val == 0)	/* successfully got a file */
			{
				if (!use_unicast)
				{
					i = sprintf(address_buf, "%d.%d.%d.%d", start_addr[0], 
							start_addr[1], start_addr[2], start_addr[3]);
				}
				/* set the data for this val */
				RegSetValueEx(hTftpdFilesKey, (LPBYTE)TmpBuf, (DWORD)0,
				 (DWORD)REG_SZ, (LPBYTE)address_buf, (DWORD)i+1);
				if (!use_unicast)
				{
					start_addr[3]++;	/* next address */
					if (start_addr[3] == 0)
					{
						start_addr[2]++;
						if (start_addr[2] == 0)
						{
							start_addr[1]++;
							if (start_addr[1] == 0)
							{
								start_addr[0]++;
								if (start_addr[0] > 239)
								{
									use_unicast = 1; 
									i = sprintf(address_buf, "%d.%d.%d.%d", 0,0,0,0); 
								}
							}
						}
					}
				}
			}	/* ret_val == 0 */
			else
			{
#ifdef _DEBUG
	DEBUGMSG(_TEXT("Enumerate failed @ index %d\r\n"), ValSize1);
#endif
			}
			ValSize1++;		/* get next value-name */
		}	/* while */
	}
	/* read the high performance byte */
	ValSize = 256;
	if (RegQueryValueEx(hTftpdKey, HIGH_PERFORMANCE, NULL, &ValType, 
						(LPBYTE)TmpBuf, &ValSize))
		/* failed to find */
		McastSegSize = DEFAULT_SEG_SIZE;
	else
	{
		if (ValSize && *(DWORD *)TmpBuf > 0L)  
			McastSegSize = MCAST_SEG_SIZE;
		else
			McastSegSize = DEFAULT_SEG_SIZE;

	}
	
	/* read the high performance byte */
	ValSize = 256;
	if (RegQueryValueEx(hTftpdKey, MTFTP_BASE_DIR_VALUE, NULL, &ValType, 
						(LPBYTE)Base_Path, &ValSize))
		/* failed to find */
		Base_Path[0] = 0;
	else
	{
		int tmp_len;
		/* translate / -> \ */ 
		q= Base_Path;
		while ((q= strchr (q, '/')) != 0)
			*q++= '\\';
		tmp_len = strlen (Base_Path);
		if (Base_Path[tmp_len-1] != '\\')
		{
			Base_Path[tmp_len] = '\\';
			Base_Path[tmp_len+1] = 0;
		}
	}

#ifdef _DEBUG
	DEBUG_CONSTS(_TEXT("MTFTPD Key processing is over\r\n"));
#endif
#endif

	/* open the log file */
	if (TFTPDLogFil) {
#if defined(WNAPP) || defined(NTAPP) || defined(NTSVC) || defined(NW) || defined(OS2)
		/* translate / -> \ */ 
		q= TFTPDLogFil;
		while ((q= strchr (q, '/')) != 0)
			*q++= '\\';
#endif

		if ((TFTPDLogFp= fopen (TFTPDLogFil, "a")) == 0) {
			log0("TFTPD[fopen()]: fn= %s, errno= %d\n",
				TFTPDLogFil, errno);
			exit (1);
		}
	}

	/* change to the TFTPD directory */
	if (TFTPDChdir) {
#if defined(WNAPP) || defined(NTAPP) || defined(NTSVC) || defined(NW) || defined(OS2)
		/* translate / -> \ */ 
		q= TFTPDChdir;
		while ((q= strchr (q, '/')) != 0)
			*q++= '\\';
#endif

		if (chdir (TFTPDChdir) < 0) {
			log0("TFTPD[chdir()]: dir= %s, errno= %d\n",
				TFTPDChdir, errno);
			exit (1);
		}
	}

	/* set default TFTPD directory if restricted */
	if (TFTPDRestrict && !TFTPDChdir)
		TFTPDChdir= DEF_TFTPDCHDIR;

	/* figure out the port to use or take default on failure */
	if (TFTPDPort == 0)
#ifdef WNAPP
		if ((sp= (struct servent far *)getservbyname
			("tftp", "udp")) == 0)
#else
		if ((sp= getservbyname("tftp", "udp")) == 0)
#endif
			TFTPDPort= htons(DEF_TFTPDPORT);
		else 
			if (sp->s_port == 0)
				TFTPDPort= htons(DEF_TFTPDPORT);
			else
#ifdef NW
				TFTPDPort= htons((short)ntohl(sp->s_port));
#else
				TFTPDPort= sp->s_port;
#endif

	/* copyright stuff */
	TFTPDVerbose = 2;
	log2("TFTPD: Starting the LCM %s \n", DESCRIPTION);
	//log1("TFTPD: %s\n", COPYRIGHT);
	TFTPDVerbose = DEF_TFTPDVERBOSE;
	/* print out configuration */
	log2("TFTPD: Chdir= %s, Restrict= %s, Exit after %d s inactivity\n",
		TFTPDChdir ? TFTPDChdir : "none",
		TFTPDRestrict ? "Yes" : "No", TFTPDTimeout);
	log2("TFTPD: Logfile= %s, Verbosity= level %d, Read ahead= %s\n",
		TFTPDLogFil ? TFTPDLogFil : "none", TFTPDVerbose,
		TFTPDAhead ? "enabled" : "disabled");
	log2("TFTPD: Maxtrans= %d TFTP transfers,"
		" Tmoutval= %d s, Write= %s\n",
		TFTPDMaxTrans, TFTPDTmoutval,
		TFTPDWrite ? (TFTPDCreat ?
			"enabled/create" : "enabled" ) : "disabled");
	log2("TFTPD: TFTP Port= %d, LTFTP Port= %d, LTFTP Segsize= %d\n",
		ntohs(TFTPDPort), ntohs(LTFTPDPort), LTFTPDSegsize);
	log2("TFTPD: MTFTP Port= %d, Unicasts= %d, MTFTP config= %s\n",
		ntohs(MTFTPDPort), TFTPDUnicast,
		TFTPDMconfig ? TFTPDMconfig : "none");

#ifdef AVL
	if (Mtftpc[0].fnam[0]) {
		i= 0;
		while (Mtftpc[i].fnam[0] && i < MTFTPDMAXFILES) {
			log2("TFTPD: MTFTP fnam= %s, MTFTP addr= %d.%d.%d.%d"
				"/%d\n",
			Mtftpc[i].fnam,
			(int)(Mtftpc[i].addr>>24) & 0xff,
			(int)(Mtftpc[i].addr>>16) & 0xff,
			(int)(Mtftpc[i].addr>> 8) & 0xff,
			(int)(Mtftpc[i].addr    ) & 0xff,
			Mtftpc[i].port);
			i++;
		}
	}
#endif	/* AVL */
	/* allocate a socket... */
	if ((TFTPDSocket= socket (AF_INET, SOCK_DGRAM, 0)) < 0)
		{ log0("TFTPD[socket()]: errno= %d\n", errno); exit (1); }

	/* ...and bind him to the port where the TFTP requests are sent */
	saddr.sin_family= AF_INET;
	saddr.sin_addr.s_addr= htonl(INADDR_ANY);
	saddr.sin_port= TFTPDPort;
	if (bind (TFTPDSocket, (SA *)&saddr, sizeof (saddr)) < 0) {
#if defined(WNAPP) || defined(NTAPP) || defined(NTSVC)
		log0("TFTPD[bind()]: errno= %d, LastError= %d\n",
			errno, WSAGetLastError());
#else
		log0("TFTPD[bind()]: errno= %d\n", errno);
#endif
		TFTPDatexit();
		exit (1);
	}

	if (!LTFTPDFlag)
		goto no_ltftpd;

	/* allocate a socket... */
	if ((LTFTPDSocket= socket (AF_INET, SOCK_DGRAM, 0)) < 0)
		{ log0("LTFTPD[socket()]: errno= %d\n", errno); exit (1); }

	/* ...and bind him to the port where the LTFTP requests are sent */
	saddr.sin_family= AF_INET;
	saddr.sin_addr.s_addr= htonl(INADDR_ANY);
	saddr.sin_port= LTFTPDPort;
	if (bind (LTFTPDSocket, (SA *)&saddr, sizeof (saddr)) < 0) {
#if defined(WNAPP) || defined(NTAPP) || defined(NTSVC)
		log0("LTFTPD[bind()]: errno= %d, LastError= %d\n",
			errno, WSAGetLastError());
#else
		log0("LTFTPD[bind()]: errno= %d", errno);
#endif
		TFTPDatexit();
		exit (1);
	}

no_ltftpd:
	if (!MTFTPDFlag)
		goto no_mtftpd;

	/* allocate a socket... */
	if ((MTFTPDSocket= socket (AF_INET, SOCK_DGRAM, 0)) < 0)
		{ log0("MTFTPD[socket()]: errno= %d\n", errno); exit (1); }

#ifdef OS2
	{
		int		on;
		struct linger	ln;

		on= 1;
		if (setsockopt (MTFTPDSocket, SOL_SOCKET,
			SO_BROADCAST, (char far *)&on, sizeof (on)) < 0) {
			log0("MTFTPD[setsockopt()]: %d\n", errno);
		}
	}
#endif

	/* ...and bind him to the port where the MTFTP requests are sent */
	saddr.sin_family= AF_INET;
	saddr.sin_addr.s_addr= htonl(INADDR_ANY);
	saddr.sin_port= MTFTPDPort;
	if (bind (MTFTPDSocket, (SA *)&saddr, sizeof (saddr)) < 0) {
#if defined(WNAPP) || defined(NTAPP) || defined(NTSVC)
		log0("MTFTPD[bind()]: errno= %d, LastError= %d\n",
			errno, WSAGetLastError());
#else
		log0("MTFTPD[bind()]: errno= %d", errno);
#endif
		TFTPDatexit();
		exit (1);
	}

	{
	int mhops = 128;
#ifdef _DEBUG
	int ttl_len;
#endif

#ifdef _DEBUG
	mhops = -1;
	ttl_len = sizeof(mhops);
	if (getsockopt (MTFTPDSocket, IPPROTO_IP, IP_MULTICAST_TTL, (char far *)&mhops, &ttl_len) < 0)
		DEBUGMSG(_TEXT("Getsockopt failed %d\r\n"),  WSAGetLastError());

	else DEBUGMSG(_TEXT("Getsockopt OK ttl=%d\r\n"), mhops);
	mhops = 128;
#endif
	if (setsockopt (MTFTPDSocket, IPPROTO_IP,
		IP_MULTICAST_TTL, (char *)&mhops, sizeof (mhops)) < 0) {
		log0("MTFTPD[setsockopt()]: %d\n", errno);
#ifdef _DEBUG
	DEBUGMSG(_TEXT("Setsockopt failed  - %d\r\n"), WSAGetLastError());
#endif
		}
	else
		{
#ifdef _DEBUG
			DEBUG_CONSTS(_TEXT("IPPROTO_IP Setsockopt successful"));

		mhops = -1;
	ttl_len = sizeof(mhops);
	if (getsockopt (MTFTPDSocket, IPPROTO_IP, IP_MULTICAST_TTL, (char far *)&mhops, &ttl_len) < 0)
		DEBUGMSG(_TEXT("Getsockopt failed %d\r\n"),  WSAGetLastError());

	else DEBUGMSG(_TEXT("Getsockopt OK again ttl=%d\r\n"), mhops);
#endif	
		}
	}


no_mtftpd:
	/* create 'semaphore' to limit number of concurrent TFTP transfers */
#ifdef NW
	atexit (TFTPDatexit);
#endif

	/* loop for TFTP requests */
#if defined(NTAPP) || defined(NTSVC)
# ifdef NTSVC
	if (!NTsvc) {
# endif
		if (TFTPDSocket >= 0) {
			if ((rc= WSAAsyncSelect (TFTPDSocket, hwndMain, WSA_ASYNC1, FD_READ)) < 0) {
				log0("WSAAsynSelect(TFTPD): %d\n", WSAGetLastError());
				TFTPDatexit();
				exit (1);
    		}
		}

		if (LTFTPDSocket >= 0) {
			if ((rc= WSAAsyncSelect (LTFTPDSocket, hwndMain, WSA_ASYNC2, FD_READ)) < 0) {
				log0("WSAAsynSelect(LTFTPD): %d\n", WSAGetLastError());
				TFTPDatexit();
				exit (1);
    		}
		}

		if (MTFTPDSocket >= 0) {
			if ((rc= WSAAsyncSelect (MTFTPDSocket, hwndMain, WSA_ASYNC3, FD_READ)) < 0) {
				log0("WSAAsynSelect(MTFTPD): %d\n", WSAGetLastError());
				TFTPDatexit();
				exit (1);
	    	}
		}

		if (!(rc= SetTimer (hwndMain, TIMER_ID, (long)TFTPDTmoutval*1000, NULL))) {
			log0("SetTimer(): %d\n", rc);
			TFTPDatexit();
			exit (1);
		}
# ifdef NTSVC
	}
	else {
# endif
#endif
#ifndef NTAPP
	for (;;) {
		/* wait for data */
		tmout.tv_sec= TFTPDTmoutval;
		tmout.tv_usec= 0;
		FD_ZERO(&rfd);
		FD_SET(TFTPDSocket, &rfd);
		if (LTFTPDSocket >= 0)
			FD_SET(LTFTPDSocket, &rfd);
		if (MTFTPDSocket >= 0)
			FD_SET(MTFTPDSocket, &rfd);
		l= TFTPDSocket;
		if (l < LTFTPDSocket)
			l= LTFTPDSocket;
		if (l < MTFTPDSocket)
			l= MTFTPDSocket;
#ifdef LINUX
		n= select (l+1, (fd_set *)&rfd, (fd_set *)0, (fd_set *)0, (struct timeval *)&tmout);
#else
		n= select (l+1, (struct fd_set *)&rfd, (struct fd_set *)0, (struct fd_set *)0, (struct timeval *)&tmout);
#endif
		if (n <= 0) {
# if defined(WNAPP) || defined(NTSVC)
			if (!doclose) {
				tcnt++;
				if (!TFTPDTimeout ||
					tcnt * TFTPDTmoutval < TFTPDTimeout) {
					tclean();
					goto retry;
				}
				log1("TFTPD: TFTP daemon inactivity timeout\n");
			}

			TFTPDatexit();
			log0("TFTPD[select()]: errno= %d, LastError= %d\n",
				errno, WSAGetLastError());
# else
			tcnt++;
			if (!TFTPDTimeout ||
				tcnt * TFTPDTmoutval < TFTPDTimeout) {
				tclean();
				goto retry;
			}
			log0("TFTPD[select()]: errno= %d\n", errno);
# endif
			exit (0);
		}

		/* handle this connection */	
		tcnt= 0;
		if (n >= 1 && FD_ISSET(TFTPDSocket, &rfd)) {
			handle(TFTPDSocket, "RRQ");
			n--;
		}

		if (n >= 1 && MTFTPDSocket >= 0 && FD_ISSET(MTFTPDSocket, &rfd)) {
			handle(MTFTPDSocket, "MRRQ");
			n--;
		}

		if (n >= 1 && LTFTPDSocket >= 0 && FD_ISSET(LTFTPDSocket, &rfd)) {
			handle(LTFTPDSocket, "LRRQ");
			n--;
		}

retry:
	;
	}
# ifdef NTSVC
	}
# endif
#endif

	return (0);
}


#if defined(NTAPP) || defined(NTSVC)
void 
/* this centers a child windows within the partent window */
CenterWnd(HWND hWnd, HWND hParentWnd, BOOL bPaint)
{
	RECT		rc2, rc1;
	RECT FAR	*lprc;
	int			nWidth, nHeight, cxCenter, cyCenter;
	
	/* if we have no parent, use desktop! */
	if (!hParentWnd)
		hParentWnd= GetDesktopWindow();
    
	GetWindowRect (hParentWnd, &rc2);
	lprc= (RECT FAR *)&rc2;
  
	cxCenter= lprc->left+((lprc->right-lprc->left)/2);
	cyCenter= lprc->top+((lprc->bottom-lprc->top)/2);

	GetWindowRect (hWnd, &rc1);
	nWidth = rc1.right-rc1.left;  
	nHeight= rc1.bottom-rc1.top;
	
	MoveWindow (hWnd, cxCenter-(nWidth/2), cyCenter-(nHeight/2), nWidth, nHeight, bPaint);    
	
	return;	
}


/* The about dialog box */
BOOL CALLBACK
Dlg_About(HWND hDlg, UINT msg, UINT wParam, LPARAM lParam)
{
	BOOL	bRet= FALSE;
	char	achDataBuf[1024];
    
	switch (msg) {
       	case WM_INITDIALOG:
			wsprintf (achDataBuf, "%s\nVersion: %s\nCopyright: %s", DESCRIPTION, TFTPD_VERSION, COPYRIGHT);
            SetDlgItemText (hDlg, DIALOG_ABOUT_INFO, (LPCSTR)achDataBuf);
			CenterWnd (hDlg, GetActiveWindow(), TRUE);
			break;
	    
		case WM_COMMAND:
    		if (wParam == DIALOG_ABOUT_OK) {
		   		EndDialog (hDlg, 0);
           		bRet= TRUE;
			}
			break;

		case WM_SYSCOMMAND:
			if (wParam == SC_CLOSE) {
           		EndDialog (hDlg, 0);
           		bRet= TRUE;
			}
			break;
	}

    return (bRet);
} 


/* Main window procedure, handles asynch messages.*/
LONG CALLBACK 
WndProc(HWND hwnd, UINT msg, WPARAM wParam, LONG lParam)
{
    int			i, nAddrSize= sizeof(SOCKADDR);
    WORD		WSAEvent, WSAErr, wCmd;
	PAINTSTRUCT	ps;
	HDC			hdc;

    switch (msg) {
		/* check for WSAAsyncSelect from one of our listen sockets */
		case WSA_ASYNC1:
		case WSA_ASYNC2:		
		case WSA_ASYNC3:
			WSAEvent= WSAGETSELECTEVENT (lParam);
			WSAErr= WSAGETSELECTERROR (lParam);

			/* check for error */
			if (WSAErr)
				break;
			
			/* check for the event */
			switch (WSAEvent) {
				case FD_READ:
					switch (msg) {
					case WSA_ASYNC1:
						Tcnt= 0;
						KillTimer (hwndMain, TIMER_ID);
						handle(TFTPDSocket, "RRQ");
						SetTimer (hwndMain, TIMER_ID, (long)TFTPDTmoutval*1000, NULL);
						break;

					case WSA_ASYNC2:
						Tcnt= 0;
						KillTimer (hwndMain, TIMER_ID);
						handle(LTFTPDSocket, "LRRQ");
						SetTimer (hwndMain, TIMER_ID, (long)TFTPDTmoutval*1000, NULL);
						break;

					case WSA_ASYNC3:
						Tcnt= 0;
						KillTimer (hwndMain, TIMER_ID);
						handle(MTFTPDSocket, "MRRQ");
						SetTimer (hwndMain, TIMER_ID, (long)TFTPDTmoutval*1000, NULL);
						break;

					default:
						break;
					}
				default:
					break;
			}
			break;
								

	case WM_TIMER:
		/* check for time outs */
		Tcnt++;
		if (!TFTPDTimeout ||
			Tcnt * TFTPDTmoutval < TFTPDTimeout) {
			tclean();
			break;
		}
		log1("TFTPD: TFTP daemon inactivity timeout\n");
		TFTPDatexit();
		exit (1);

		/* !UNREACHED! */
		break;

	case WM_COMMAND:
#ifdef WIN32
		wCmd= LOWORD(wParam);
#else
		wCmd= (WORD)wParam;
#endif
        switch (wCmd) {

			case MENU_HELP_ABOUT:
				DialogBox (hInst, MAKEINTRESOURCE(DIALOG_ABOUT), hwnd, Dlg_About);
				break;
               
			case MENU_FILE_EXIT:
				PostMessage (hwnd, WM_CLOSE, 0, 0);
				break;
                       
           	default:
				return (DefWindowProc (hwnd, msg, wParam, lParam));
			} 
		break;

	case WM_QUERYENDSESSION:
	case WM_CLOSE:
		/* Close the socket before we leave. */
        /* if (!SendMessage(hwnd, WM_COMMAND, IDM_CLOSE, 1L)) */
		DestroyWindow (hwnd);
		break;

    case WM_DESTROY:
        PostQuitMessage (0);
        break;
		
	case WM_CREATE:
        /* center dialog box */
        CenterWnd (hwnd, NULL, TRUE);
        break;

	case WM_PAINT:
		hdc= BeginPaint (hwndMain, &ps);
		SelectObject (hdc, GetStockObject (SYSTEM_FIXED_FONT));
		for (i= 0; i < YLEN; i++)
			TextOut (hdc, CHAR_WIDTH/2, CHAR_HIGHT/2 + i * CHAR_HIGHT, Screen[i], XLEN);
		EndPaint (hwndMain, &ps);
		break;

    default:
		return (DefWindowProc (hwnd, msg, wParam, lParam));   
	}

	/* !UNREACHED! */
	return (0);
}



/* Windows Application start function */
int WINAPI
ApplWinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpszCmdLine, int nCmdShow)
{
	int 		i, r;
	MSG			msg;
	WNDCLASS	wndclass;
	WSADATA		stWSAData;
	PAINTSTRUCT	ps;
	int			argc;
	char		*p, *argv[16];


	/* start output at first line */
	YLine= 0;

	/* save instance handle */
	hInst= hInstance;

	/* register window class */
	if (!hPrevInstance) {
		wndclass.style= CS_HREDRAW | CS_VREDRAW;
      	wndclass.lpfnWndProc= WndProc;
      	wndclass.cbClsExtra= 0;
      	wndclass.cbWndExtra= 0;
      	wndclass.hInstance= hInstance;
      	wndclass.hIcon= LoadIcon(hInst, MAKEINTRESOURCE(ICON_TFTPD));
      	wndclass.hCursor= LoadCursor(NULL,IDC_ARROW);
      	wndclass.hbrBackground= (HBRUSH)(COLOR_WINDOW+1);
      	wndclass.lpszMenuName= MAKEINTRESOURCE(MENU_TFTPD);
      	wndclass.lpszClassName= "TFTPD";
       
      	if (!RegisterClass (&wndclass)) {
        	return (0);
      	}
    }

	hwndMain= CreateWindow (
		SZSERVICENAME,
		DESCRIPTION,
		WS_OVERLAPPEDWINDOW,
		CW_USEDEFAULT,
		CW_USEDEFAULT,
		VXLEN*CHAR_WIDTH,
		(4+YLEN)*CHAR_HIGHT,
		NULL,
		NULL,
		hInstance,
		NULL
	);

	/* exit if class registration failed */
	if (!hwndMain)
		return (0);

	/* display our window */
	ShowWindow (hwndMain, nCmdShow);
	UpdateWindow (hwndMain);

	/* initialize text output */
	InvalidateRect (hwndMain, NULL, TRUE);
	Hdc= BeginPaint (hwndMain, &ps);
	SelectObject (Hdc, GetStockObject (SYSTEM_FIXED_FONT));
	memset (Screen, ' ', sizeof (Screen));

	log0("12345678901234567890123456789012345678901234567890123456789012345678901234567890");
	log0("********************************************************************************");
	log0("**********                   THIS IS A BETA VERSION                   **********");
	log0("********************************************************************************");
	/* initialize WinSock DLL */
	r= WSAStartup (MAKEWORD(2, 0), &stWSAData);
    if (r != 0) {    
  		log0("WSAStartup() failed");
		exit (1);
    }
	else {
	    /* break lpszCmdLine into argc/argv */
		p= lpszCmdLine;
		argc= 1;
		argv[0]= malloc(80);
		argv[0][0]= 0;

		while (*p) {
			argv[argc]= malloc(80);
			while (*p && *p == ' ')
				p++;
			i= 0;
			while (*p && *p != ' ') {
				argv[argc][i]= *p;
				p++;
				i++;
			}
			argv[argc][i]= 0;
			if (i)
				argc++;
		}
		argv[argc]= 0;

	    /* now call TFTP server */
		Mmain(argc, argv);
    
		/* main message loop */
		while (GetMessage (&msg, NULL, 0, 0)) {
			TranslateMessage (&msg);
			DispatchMessage (&msg);
		}	
	}

	return (0);
}
#endif


#if defined(NTAPP) && !defined(NTSVC)
int WINAPI
WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpszCmdLine, int nCmdShow)
{
	return (ApplWinMain(hInstance, hPrevInstance, lpszCmdLine, nCmdShow));
}
#endif


#ifdef NTSVC
/* this reports the current status to the SCM */
BOOL 
ReportStatusToSCMgr(DWORD dwCurrentState,
	DWORD dwWin32ExitCode, DWORD dwWaitHint)
{
    static DWORD	dwCheckPoint= 1;
    BOOL			fResult= TRUE;


	/* when debugging we don't report to the SCM */
    if (NTsvc) {
        if (dwCurrentState == SERVICE_START_PENDING)
            ssStatus.dwControlsAccepted= 0;
        else
            ssStatus.dwControlsAccepted= SERVICE_ACCEPT_STOP;

        ssStatus.dwCurrentState= dwCurrentState;
        ssStatus.dwWin32ExitCode= dwWin32ExitCode;
        ssStatus.dwWaitHint= dwWaitHint;

        if ((dwCurrentState == SERVICE_RUNNING) ||
             (dwCurrentState == SERVICE_STOPPED))
            ssStatus.dwCheckPoint= 0;
        else
            ssStatus.dwCheckPoint= dwCheckPoint++;


        /* report the status of the service to the service control manager */
        //
        if (!(fResult= SetServiceStatus( sshStatusHandle, &ssStatus))) {
            log0("SetServiceStatus failed %s", SZSERVICENAME);
			fResult= FALSE;
		}
    }
    return (fResult);
}


/* service control routine */
VOID WINAPI 
service_ctrl(DWORD dwCtrlCode)
{
    /* handle the requested control code */
    switch(dwCtrlCode) {
        /* stop the service */
        case SERVICE_CONTROL_STOP:
			ReportStatusToSCMgr(SERVICE_STOP_PENDING, NO_ERROR, 0);
			ssStatus.dwCurrentState= SERVICE_STOPPED;

			StopErrno = TRUE;
//AVL			WSACancelBlockingCall (); // problem with wsock 2.0
			TFTPDatexit();
			TFTPDVerbose = 2;
			log2("Stopped %s\n", SZSERVICENAME);
			TFTPDVerbose = DEF_TFTPDVERBOSE;

 			break;

        /* Update the service status */
        case SERVICE_CONTROL_INTERROGATE:
            break;

        /* invalid control code */
        default:
            break;
    }

    ReportStatusToSCMgr(ssStatus.dwCurrentState, NO_ERROR, 0);
}



/* main if this is a NT service */
void WINAPI
service_main(DWORD dwArgc, LPSTR *lpszArgv)
{
	WSADATA		stWSAData;
	int			i, argc;
	char		*p, *argv[16];
	
	/* register our service control handler */
    sshStatusHandle= RegisterServiceCtrlHandler (SZSERVICENAME, service_ctrl);
	if (!sshStatusHandle) {
		log0("Can't register service %s", SZSERVICENAME);
        goto cleanup;
	}

    /* SERVICE_STATUS members that don't change */
    ssStatus.dwServiceType= SERVICE_WIN32_OWN_PROCESS;
    ssStatus.dwServiceSpecificExitCode= 0;

    /* report the status to the service control manager.*/
    ReportStatusToSCMgr(SERVICE_RUNNING, NO_ERROR, 3000);

	if (WSAStartup (MAKEWORD(2, 0), &stWSAData) != 0) {
  		log0("WSAStartup() failed");
		goto cleanup;
    }

	/* break CmdLine into argc/argv */
	p= Cmdline;
	argc= 1;
	argv[0]= malloc(80);
	argv[0][0]= 0;

	while (*p) {
		argv[argc]= malloc(80);
		while (*p && *p == ' ')
			p++;
		i= 0;
		while (*p && *p != ' ') {
			argv[argc][i]= *p;
			p++;
			i++;
		}
		argv[argc][i]= 0;
		if (i)
			argc++;
	}
	argv[argc]= 0;
	
	Mmain(argc, argv);

cleanup:

    /* try to report the stopped status to the service control manager */
    if (sshStatusHandle)
        ReportStatusToSCMgr (SERVICE_STOPPED, dwErr, 0);

    return;
}


/* copy last error into string */
LPTSTR 
GetLastErrorText(LPTSTR lpszBuf, DWORD dwSize)
{
    DWORD	dwRet;
    LPTSTR	lpszTemp= NULL;

    dwRet= FormatMessage (
		FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM |FORMAT_MESSAGE_ARGUMENT_ARRAY,
        NULL,
        GetLastError(),
        LANG_NEUTRAL,
        (LPTSTR)&lpszTemp,
        0,
        NULL);

    /* supplied buffer is not long enough */
    if (!dwRet || ((long)dwSize < (long)dwRet+14))
        lpszBuf[0]= 0;
	else {
		/* remove cr and newline character */
        lpszTemp[lstrlen(lpszTemp)-2]= 0;  
        sprintf (lpszBuf, "%s (0x%x)", lpszTemp, GetLastError ());
    }

    if (lpszTemp)
        LocalFree ((HLOCAL)lpszTemp);

    return (lpszBuf);
}


/* installs the service */
int
CmdInstallService()
{
    SC_HANDLE   schService;
    SC_HANDLE   schSCManager;
    TCHAR		szPath[512];
	int			ret;

	ret= 0;
    if (GetModuleFileName (NULL, szPath, 512) == 0) {
        log0("Unable to install %s - %s\n", SZSERVICENAME, GetLastErrorText (szErr, 256));
		return (1);
    }

    schSCManager= OpenSCManager (NULL, NULL, SC_MANAGER_ALL_ACCESS);
    if (schSCManager) {
        schService= CreateService (
            schSCManager,               /* SCManager database */
            SZSERVICENAME,				/* name of service */
            DESCRIPTION,				/* name to display */
            SERVICE_ALL_ACCESS,         /* desired access */
            SERVICE_WIN32_OWN_PROCESS,  /* service type */
            SERVICE_AUTO_START,         /* start type */
            SERVICE_ERROR_NORMAL,       /* error control type */
            szPath,                     /* service's binary */
            NULL,                       /* no load ordering group */
            NULL,                       /* no tag identifier */
			SZDEPENDENCIES,				/* dependencies */
            //"",							/* dependencies */
            NULL,                       /* LocalSystem account */
            NULL);                      /* no password */

        if (schService) 
		{
			TFTPDVerbose = 2;
            log2("Installed %s\n", SZSERVICENAME);
			TFTPDVerbose = DEF_TFTPDVERBOSE;
            CloseServiceHandle (schService);
        }
		else {
            log0("xCreateService failed %s - %s\n", SZSERVICENAME, GetLastErrorText (szErr, 256));
			ret= 1;
		}
       
        CloseServiceHandle (schSCManager);
    }
    else {
		log0("OpenSCManager failed %s - %s\n", SZSERVICENAME, GetLastErrorText (szErr,256));
		ret= 1;
	}

	return (ret);
}


/* stops and removes the service */
int
CmdRemoveService()
{
	SC_HANDLE   schService;
    SC_HANDLE   schSCManager;
	int			ret;

	ret= 0;
    schSCManager= OpenSCManager (NULL, NULL, SC_MANAGER_ALL_ACCESS);
    if (schSCManager) {
        schService= OpenService (schSCManager, SZSERVICENAME, SERVICE_ALL_ACCESS);
        if (schService) {
            /* try to stop the service */
            if (ControlService (schService, SERVICE_CONTROL_STOP, &ssStatus)) {
				Sleep (1000);
				while (QueryServiceStatus (schService, &ssStatus)) {
                    if (ssStatus.dwCurrentState == SERVICE_STOP_PENDING)
						Sleep (1000);
              		else
                        break;
                }

                if (ssStatus.dwCurrentState == SERVICE_STOPPED)
				{
                    log2("Stopped %s\n", SZSERVICENAME); // this is not present in the log 
														// because the Verbose mode is set to 1
				}
                else 
				{
                    log0("Failed to stop %s\n", SZSERVICENAME);
					ret= 1;
				}
            }

            // now remove the service
            if (DeleteService (schService))
			{
				TFTPDVerbose = 2;
                log2("Removed %s\n", SZSERVICENAME);
				TFTPDVerbose = DEF_TFTPDVERBOSE;
			}
            else 
			{
                log0("DeleteService failed %s - %s\n", 
					SZSERVICENAME, GetLastErrorText (szErr,256));
				ret= 1;
			}

            CloseServiceHandle (schService);
        }
        else {
            log0("OpenService failed %s - %s\n", 
				SZSERVICENAME, GetLastErrorText (szErr,256));
			ret= 1;
		}

        CloseServiceHandle (schSCManager);
    }
    else {
        log0("OpenSCManager failed %s - %s\n", 
			SZSERVICENAME, GetLastErrorText (szErr,256));
		ret= 1;
	}

	return (ret);
}


int WINAPI
WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpszCmdLine, int nCmdShow)
{
	FILE	*fp;

	SERVICE_TABLE_ENTRY dispatchTable[]= {
		{ TEXT(SZSERVICENAME), (LPSERVICE_MAIN_FUNCTION)service_main },
		{ NULL, NULL }
	};

	/* running as NT service */
	NTsvc= 1;

	/* if no arguments are passed, read in /etc/tftptab */
	if (lpszCmdLine == 0 || (lpszCmdLine && *lpszCmdLine == 0)) {
		if ((fp= fopen (DEF_TFTPTAB, "r")) != NULL) {
			fgets (Cmdline, sizeof(Cmdline), fp);
			fclose (fp);
		}
	}
	else
		strncpy (Cmdline, lpszCmdLine, sizeof(Cmdline));

	/* check for basic options */
	TFTPDVerbose= DEF_TFTPDVERBOSE;
	if (!strncmp (Cmdline, "-install", 8))
		return (CmdInstallService());
	
	if (!strncmp (Cmdline, "-remove", 7)) 
		return (CmdRemoveService());

	if (!strncmp (Cmdline, "-cmd", 4)) {
		NTsvc= 0;
		return (ApplWinMain(hInstance, hPrevInstance, Cmdline+4, nCmdShow));
	}

	if (!strncmp (Cmdline, "-?", 2) || !strncmp (Cmdline, "/?", 2)) {
		NTsvc= 0;
		return (ApplWinMain(hInstance, hPrevInstance, Cmdline, nCmdShow));
	}


	/* call the dispatcher */
	if (!StartServiceCtrlDispatcher (dispatchTable))
		log0("StartServiceCtrlDispatcher failed %s", SZSERVICENAME);

	return (0);
}
#endif
